// Copyright (C) 2009  Philip Ashmore (contact@philipashmore.com)
// License: LGPLv3.  See LICENSE.txt for the full license.

/** \file
 * \brief Source for v3c::cmdline::Parser
 */

#include "parser.hpp"
#include <v3c/trace_scope.hpp>
#include <v3c/string_utils.hpp>

namespace v3c {
namespace cmdline {

FlexxTokenPtr Scanner::make_token(int t)
{
	return new FlexxToken(t, *this, yyleng);
}

void Parser::parse(const char * shortopts_, const Option * longopts_)
{
	TraceWrap(__PRETTY_FUNCTION__);

	m_szShortOpts = shortopts_;
	m_pLongOpts = longopts_;
	parse();
	if(m_bNeedParam)
		on_opt_arg(m_nLastArg, m_sValuePart.c_str());
	else if(m_bValueAssigned)
		on_arg(m_sValuePart.c_str());
}
void Parser::on_short_arg()
{
	const char * text = m_scanner.YYText();
	bool have_param = strlen(text) > 2;
	TraceWrap(__PRETTY_FUNCTION__ << " : " << text);

	if(m_bNeedParam) {
		trace << "Supplying as parameter value." << endl;
		on_opt_arg(m_nLastArg, text);
		arg_reset();
		return;
	}
	// See if we're expecting this parameter.
	char c = text[1];
	const char *p = strchr(m_szShortOpts, c);
	if(!p)
		ABORT("Unexpected option '%c'", c);
	m_bNeedParam = (p[1] == ':');
	if(!have_param) {
		if(m_bNeedParam)
			m_nLastArg = c;
		else{
			on_opt(text[1]);
		}
	}else{
		if(m_bNeedParam) {
			on_opt_arg(c, text + 2);
			m_bNeedParam = false;
		}else
			ABORT("Unexpected parameter \"%s\" supplied to '%c'", text + 2, c);
	}
}
void Parser::on_long_arg()
{
	const char * text = m_scanner.YYText();
	TraceWrap(__PRETTY_FUNCTION__ << " : " << text);
	// Find the table entry.
	const Option * p = m_pLongOpts;
	for(; p->name; ++p)
		if(!strcmp(text + 2, p->name))
			break;
	if(!p->name)
		ABORT("Long option \"%s\" not found.", text);
	m_bNeedParam = (p->arg == Option::eRequired);
	if(m_bNeedParam)
		m_nLastArg = p->val;
}
void Parser::on_long_arg_equals()
{
	const char * text = m_scanner.YYText();
	TraceWrap(__PRETTY_FUNCTION__ << " : " << text);
	// Find the table entry.
	const Option * p = m_pLongOpts;
	for(; p->name; ++p)
		if(!strcmp(text + 2, p->name))
			break;
	if(!p->name)
		ABORT("Long option \"%s\" not found.", text);
	m_bNeedParam = (p->arg == Option::eRequired);
	if(!m_bNeedParam)
		ABORT("Long option \"%s\" does not take a parameter.", text);
	m_nLastArg = p->val;
}
void Parser::on_value()
{
	const char * text = m_scanner.YYText();
	TraceWrap(__PRETTY_FUNCTION__ << " : " << text);
	// This may only be part of the value.
	m_bValueAssigned = true;
	m_sValuePart += text;
}
void Parser::on_double_quoted_string()
{
	const char * text = m_scanner.YYText();
	TraceWrap(__PRETTY_FUNCTION__ << " : " << text);
	m_sValuePart += string(text+ 1, strlen(text)-2);
	if(m_bNeedParam) {
		on_opt_arg(m_nLastArg, m_sValuePart.c_str());
		arg_reset();
	}
}
void Parser::on_single_quoted_string()
{
	const char * text = m_scanner.YYText();
	TraceWrap(__PRETTY_FUNCTION__ << " : " << text);
	m_sValuePart += string(text+ 1, strlen(text)-2);
	if(m_bNeedParam) {
		on_opt_arg(m_nLastArg, m_sValuePart.c_str());
		arg_reset();
	}
}
void Parser::on_space()
{
	TraceWrap(__PRETTY_FUNCTION__);
	if(m_nLastArg == -1) {
		if(m_bValueAssigned) {
			on_arg(m_sValuePart.c_str());
			arg_reset();
		}
		return;
	}
	if(m_bNeedParam && !m_bValueAssigned) {
		trace << "Parameter to follow." << endl;
		return;
	}
	on_opt_arg(m_nLastArg, m_sValuePart.c_str());
	arg_reset();
}
void Parser::arg_reset()
{
	m_sValuePart = "";
	m_bNeedParam = false;
	m_bValueAssigned = false;
	m_nLastArg = -1;
}
void Parser::on_non_positional_value()
{
	TraceWrap(__PRETTY_FUNCTION__);
	on_arg(m_scanner.YYText());
}
/// Any paramaters that look like arguments from now on are to be treated as
/// regular parameters.
void Parser::on_positional_end()
{
	TraceWrap(__PRETTY_FUNCTION__);
}

void Parser::ABORT(const char * fmt, ...)
{
	string err;
	va_list args;
	va_start(args, fmt);
	bool bRet = string_vprintf(err, fmt, args);
	va_end(args);
	if(!bRet)
		throw runtime_error(string("Parser::ABORT : error processing \"")
			+ fmt + "\"");
	throw runtime_error(err);
}

//virtual bool on_opt(int c) = 0;
//virtual bool on_opt_arg(int c, const char * param) = 0;
//virtual bool on_arg(const char * param) = 0;

} // namespace cmdline
} // namespace v3c
