#if !defined(v3c_flexx_hpp_included)
#define v3c_flexx_hpp_included

// Copyright (C) 2009  Philip Ashmore (contact@philipashmore.com)
// License: LGPLv3.  See LICENSE.txt for the full license.

#include <iosfwd>
#include <cstdio> // EOF
#include <stdexcept>
#include <v3c/span.hpp>
#include <v3c/smart_ptr.hpp>

namespace v3c {

using namespace std;

class FlexxToken : public Span
{
public:
	inline FlexxToken(int id_, const Position & p, unsigned int length)
	: Span(p, length)
	, m_id(id_)
	{
	}
	inline FlexxToken(int id_, const Span & s)
	: Span(s)
	, m_id(id_)
	{
	}
	inline int id() const { return m_id; }
protected:
	int m_id;
};
typedef smart_ptr< FlexxToken, SingleThreadModelTraits > FlexxTokenPtr;

struct yy_buffer_state : public Position
{
	yy_buffer_state(const Position & p, unsigned int size);
	~yy_buffer_state();
	std::istream* yy_input_file;

	char *yy_ch_buf;		/* input buffer */
	char *yy_buf_pos;		/* current position in input buffer */

	/* Size of input buffer in bytes, not including room for EOB
	 * characters.
	 */
	unsigned int yy_buf_size;

	/* Number of characters read into yy_ch_buf, not including EOB
	 * characters.
	 */
	int yy_n_chars;

	/* Whether we "own" the buffer - i.e., we know we created it,
	 * and can realloc() it to grow it, and should free() it to
	 * delete it.
	 */
	int yy_is_our_buffer;

	/* Whether this is an "interactive" input source; if so, and
	 * if we're using stdio for input, then we want to use getc()
	 * instead of fread(), to make sure we stop fetching input after
	 * each newline.
	 */
	int yy_is_interactive;

	/* Whether we're considered to be at the beginning of a line.
	 * If so, '^' rules will be active on the next match, otherwise
	 * not.
	 */
	int yy_at_bol;

	//int yy_bs_lineno; /**< The line count. */
	//int yy_bs_column; /**< The column count. */
    
	/* Whether to try to fill the input buffer when we reach the
	 * end of it.
	 */
	int yy_fill_buffer;

	int yy_buffer_status;
private:
	yy_buffer_state(const yy_buffer_state &);
	yy_buffer_state & operator =(const yy_buffer_state &);
};
typedef yy_buffer_state *YY_BUFFER_STATE;
//#define YY_BUFFER_NEW 0
//#define YY_BUFFER_NORMAL 1
	/* When an EOF's been seen but there's still some text to process
	 * then we mark the buffer as YY_EOF_PENDING, to indicate that we
	 * shouldn't try reading from the input source any more.  We might
	 * still have a bunch of tokens to match, though, because of
	 * possible backing-up.
	 *
	 * When we actually see the EOF, we change the status to "new"
	 * (via yyrestart()), so that the user can continue scanning by
	 * just pointing yyin at a new input file.
	 */
//#define YY_BUFFER_EOF_PENDING 2

class scanner_error : public runtime_error
{
public:
	inline scanner_error(const string & msg) throw()
	: runtime_error(msg)
	{
	}
};

class scanner_result : public exception
{
public:
	scanner_result(int r) throw()
	: exception()
	, result(r)
	{}
	int result;
};
/*! scanner_stage : allows for scanner-related function calls inside of scanner
member functions.
In order to provide as much compatibility as possible, existing flex actions
need only make the following modifications.
BEGIN x -> begin(x)
ECHO     -> echo()
YYSTATE  -> state()
*/
class scanner_stage : public exception
{
public:
	enum Stage
	{ goto_match, goto_find_action, goto_find_rule, goto_do_action };
	scanner_stage(Stage s) throw()
	: exception()
	, stage(s)
	{}
	Stage stage;
};

class Flexx : public Position
{
public:
	enum { YY_BUFFER_NEW, YY_BUFFER_NORMAL, YY_BUFFER_EOF_PENDING };
	enum { EOB_ACT_CONTINUE_SCAN, EOB_ACT_END_OF_FILE, EOB_ACT_LAST_MATCH };

	/* Returned upon end-of-file. */
	enum { YY_NULL = 0 };
// class FlexLexer
	Flexx(unsigned int spaces_per_tab = 4);
	// arg_yyin and arg_yyout default to the cin and cout, but we
	// only make that assignment when initializing in read().
	Flexx(istream* arg_yyin, ostream* arg_yyout, unsigned int spaces_per_tab = 4);
	virtual ~Flexx();
private:
	Flexx(const Flexx &);
	Flexx & operator=(const Flexx &);
public:
	const char* YYText() const { return yytext; }
	inline int YYLeng() const { return yyleng; }

	inline int lineno() const { return line; }

	inline int debug() const		{ return yy_flex_debug; }
	inline void set_debug( int flag )	{ yy_flex_debug = flag; }

	// This is WRONG - there's no way to generalize the interface where any
	// type could be used as the semantic value.
	//virtual int yylex() = 0;
protected:
	char* yytext;
	int yyleng;
	//int yylineno;		// only maintained if you use %option yylineno
	int yy_flex_debug;	// only has effect with -d or "%option debug"

	virtual void before_rule() {}
	virtual void after_rule() { update_position(yytext, yyleng); }

public:
	virtual void init();
// class yyFlexLexer
public:
	//void yy_switch_to_buffer(yy_buffer_state * new_buffer );
	v3c::yy_buffer_state * yy_create_buffer(istream * s, int size);
	void yy_delete_buffer(yy_buffer_state * b);
	virtual void yyrestart(istream* s);

	virtual void yypush_buffer_state( v3c::yy_buffer_state* new_buffer ) = 0;
	virtual void yypop_buffer_state() = 0;

	// Switch to new input/output streams.  A nil stream pointer
	// indicates "keep the current one".
	virtual void switch_streams(istream* new_in, ostream * new_out = 0 );
	virtual int yywrap();

protected:
	virtual int LexerInput(char* buf, int max_size);
	virtual void LexerOutput(const char* buf, int size);
	virtual void LexerError(const char* fmt,...)
		__attribute__ ((__format__ (__printf__, 2, 3)));
	virtual inline unsigned int start_stack_increment() const
	{ return 25; }

	void yyunput(int c, char* buf_ptr);
	int yyinput();

	void yy_load_buffer_state();
	void yy_init_buffer(yy_buffer_state * b, istream * s);
	void yy_flush_buffer(yy_buffer_state * b);

	int yy_start_stack_ptr;
	int yy_start_stack_depth;
	int* yy_start_stack;

	void yy_push_state( int new_state );
	void yy_pop_state();
	int yy_top_state();

	// These access DFA-specific tables so they can't be here.
	//int yy_get_previous_state();
	//int yy_try_NUL_trans( int current_state );

	int yy_get_next_buffer();

	istream* yyin;	// input source for default LexerInput
	ostream* yyout;	// output sink for default LexerOutput

	// yy_hold_char holds the character lost when yytext is formed.
	char yy_hold_char;

	// Number of characters read into yy_ch_buf.
	int yy_n_chars;

	// Points to current character in buffer.
	char* yy_c_buf_p;

	int yy_init;		// whether we need to initialize
	int yy_start;		// start state number

	// Flag which is used to allow yywrap()'s to do buffer switches
	// instead of setting up a fresh yyin.  A bit of a hack ...
	int yy_did_buffer_switch_on_eof;


	size_t yy_buffer_stack_top; /**< index of top of stack. */
	size_t yy_buffer_stack_max; /**< capacity of stack. */
	struct yy_buffer_state ** yy_buffer_stack; /**< Stack as an array. */
	virtual void yyensure_buffer_stack() = 0;

	// The following are not always needed, but may be depending
	// on use of certain flex features (like REJECT or yymore()).

	int yy_last_accepting_state;
	char* yy_last_accepting_cpos;

	int* yy_state_buf;
	int* yy_state_ptr;

	char* yy_full_match;
	int* yy_full_state;
	int yy_full_lp;

	int yy_lp;
	int yy_looking_for_trail_begin;

	int yy_more_flag;
	int yy_more_len;
	int yy_more_offset;
	int yy_prev_more_offset;

// Macros become inlines
	/* We provide macros for accessing buffer states in case in the
	* future we want to put the buffer states in a more general
	* "scanner state".
	*
	* Returns the top of the stack, or NULL.
	*/
	yy_buffer_state * yy_CurrentBuffer()
	{ return yy_buffer_stack ? yy_buffer_stack[yy_buffer_stack_top] : NULL; }

	/* Same as previous macro, but useful when we know that the buffer stack is
	* not NULL or when we need an lvalue. For internal use only.
	*/
	inline yy_buffer_state *& yy_CurrentBufferLvalue()
	{ return yy_buffer_stack[yy_buffer_stack_top]; }
	/* Promotes a possibly negative, possibly signed char to an unsigned
	* integer for use as an array index.  If the signed char is negative,
	* we want to instead treat it as an 8-bit unsigned char, hence the
	* double cast.
	*/
	inline static unsigned int promote(char c)
	{ return (unsigned int)(unsigned char) c; }
	// Enter a start condition.
	inline void begin(int s)
	{ yy_start = 1 + 2 * s; }
	/* Translate the current start state into a value that can be later handed
	 * to begin to return to the state.
	 */
	inline int state() const { return (yy_start - 1) / 2; }

// Default values become data members
	/* Number of entries by which start-condition stack grows. */
	unsigned int m_uStartStackIncr; // 25
	bool m_bInteractive;

// Possibly implemented in UserClass.
	// void yyterminate() { throw Result(END); }
	// void before_rule(); // YY_USER_ACTION
	// int yy_get_previous_state();

	enum { eTerminateValue = 0 }; // Derived class is free to replace this.
	inline void yy_do_before_action()
	{
		yytext = yy_bp;
		// %% [2.0] code to fiddle yytext and yyleng for yymore() goes here
		yytext -= yy_more_len;
		yyleng = (size_t) (yy_cp - yytext);
		yy_hold_char = *yy_cp;
		*yy_cp = '\0';
		// %% [3.0] code to copy yytext to yytext[] goes here, if %array
		yy_c_buf_p = yy_cp;
	}
	inline void yy_less_lineno(int n)
	{
		int yyl;
		for ( yyl = n; yyl < yyleng; ++yyl )
			if ( yytext[yyl] == '\n' )
				--line;
	}
	inline void accept(int t) { throw scanner_result(t); }
	inline void more() { yy_more_flag = 1; }
	inline void less(int n)
	{
		/* Undo effects of setting up yytext. */
		yy_less_lineno(n);
		*yy_cp = yy_hold_char;
		yy_c_buf_p = yy_cp = yy_bp + n - yy_more_len;
		yy_do_before_action(); // set up yytext again
	}

	//inline void unput(char c);
	//inline int input();
	virtual void yy_switch_to_buffer(v3c::yy_buffer_state * new_buffer) = 0;
	inline void echo() { LexerOutput( yytext, yyleng ); }

	virtual void report_state() = 0;
	virtual void report_state_pre() {}
	virtual void report_state_post() {}

	inline int spaces_per_tab() const { return m_spaces_per_tab; }
	inline void set_spaces_per_tab(int n) { m_spaces_per_tab = n; }

	class RuleScope
	{
	public:
		inline RuleScope(Flexx & f_)
		: f(f_)
		{
			f.before_rule();
		}
		inline ~RuleScope()
		{
			f.after_rule();
		}
	protected:
		Flexx & f;
	};

	// Were local variables.
	int yy_act; // Needed by report_state(), less().
	int yy_current_state;
	char *yy_cp, *yy_bp;
	unsigned int m_spaces_per_tab;
};
#define FLEXX_IMPL(S) \
public: \
	void report_state(); \
	void terminate() { throw v3c::scanner_result(eTerminateValue); } \
protected: \
	void unput(char c); \
	void unput(int, char *); \
	int input(); \
	void check_input(char * buf, int & result, unsigned int max_size) \
	{ if ( (result = LexerInput( buf, max_size )) < 0 ) \
		LexerError("input in flex scanner failed"); \
	} \
	int yy_get_previous_state(); \
	int yy_try_NUL_trans(int current_state); \
	void yyrestart(istream* s); \
	int yy_get_next_buffer(); \
	void yy_switch_to_buffer(v3c::yy_buffer_state * new_buffer); \
	void yy_load_buffer_state(); \
	v3c::yy_buffer_state* yy_create_buffer(istream*, int); \
	void yy_delete_buffer(v3c::yy_buffer_state*); \
	void yy_init_buffer(v3c::yy_buffer_state*, istream*); \
	void yy_flush_buffer(v3c::yy_buffer_state*); \
	void yypush_buffer_state( v3c::yy_buffer_state* new_buffer ); \
	void yypop_buffer_state(); \
	void yyensure_buffer_stack(); \
	void yy_push_state(int); \
	void yy_pop_state(); \
	int yy_top_state(); \
	static const char * m_state_names[]; \
	inline int eof_state(int state) const; \
	static const unsigned int m_guNumStates; \
public: \
	inline const char * state_name() const \
	{ return m_state_names[state()]; } \
	/* Call read with new input/output sources. */ \
	inline S read(istream* new_in, ostream* new_out = 0 ) \
	{ \
		switch_streams(new_in, new_out); \
		return read(); \
	} \
	virtual S read()

} // namespace v3c

////////////////////////////////////////////////////////////////////////////////

#endif // #if !defined(v3c_flexx_hpp_included)
