// Copyright (C) 2009  Philip Ashmore (contact@philipashmore.com)
// License: LGPLv3.  See LICENSE.txt for the full license.

#include <iostream>
#include <sstream>
#include "calc_parser.hpp"
#include <v3c/pretty_print_streambuf.hpp>
using namespace std;
using namespace v3c;

namespace calc {

class IntParser : public Parser
{
public:
	IntParser()
	: Parser()
	, variables()
	, m_result()
	{
	}
public:
	int result() const { return m_result; }
protected:
	void assign(TokenPtr & lhs, TokenPtr & rhs)
	{
		variables[lhs->string_value()] = rhs->int_value();
	}
	void store_result(TokenPtr & p)
	{
		m_result = p->int_value();
	}
	TokenPtr add(TokenPtr &a, TokenPtr &b)
	{
		Token * ret = new IntToken(*a, a->int_value() + b->int_value());
		ret->end = b->end;
		return ret;
	}
	TokenPtr sub(TokenPtr &a, TokenPtr &b)
	{
		Token * ret = new IntToken(*a, a->int_value() - b->int_value());
		ret->end = b->end;
		return ret;
	}
	TokenPtr mult(TokenPtr &a, TokenPtr &b)
	{
		Token * ret = new IntToken(*a, a->int_value() * b->int_value());
		ret->end = b->end;
		return ret;
	}
	TokenPtr div(TokenPtr &a, TokenPtr &b)
	{
		int ia = a->int_value(), ib = b->int_value();
		if(!ia)
			error(*b, "Division by zero");
		Token * ret = new IntToken(*a, ia/ib);
		ret->end = b->end;
		return ret;
	}
	TokenPtr fetch(TokenPtr &a)
	{
		const string & key = a->string_value();
		map<std::string, int>::iterator i = variables.find(key);
		if(i == variables.end()) {
			stringstream ss;
			ss << "variable " << a->string_value() << " is not defined.";
			error(*a, ss.str());
		}
		return new IntToken(*a, i->second);
	}
	map<std::string, int> variables;
	int m_result;
};

class OpToken : public Token
{
public:
	enum { eOp = Token::eNext, eNext };
	/// Called directly
	inline OpToken(int id_, TokenPtr &a, TokenPtr &b)
	: Token(eOp, id_, v3c::Span(a->begin, b->end))
	, left(a)
	, right(b)
	{
	}
	virtual inline const char * name() const
	{
		if(type() != eOp)
			return Token::name();
		switch(m_id) {
		case '+':
			return "+";
		case '-':
			return "-";
		case '*':
			return "*";
		case '/':
			return "/";
		default:
			return "?";
		}
	}
	TokenPtr left, right;
protected:
	static void stream(PrettyPrintStreambuf & buf, const Token * p)
	{
		ostream & s(buf.stream());
		switch(p->type()) {
		case eOp: {
			const OpToken * o = static_cast< const OpToken *>(p);
			if(o->left) {
				PrettyPrintStreambuf::indenter i(buf);
				OpToken::stream(buf, &*o->left); s << endl;
			}else
				s << "<null>";
			s << (char)o->id() << endl;
			if(o->right) {
				PrettyPrintStreambuf::indenter i(buf);
				OpToken::stream(buf, &*o->right); s << flush;
			}else
				s << "<null>";
			s << flush;
			break; }
		case eInt:
		case eString:
			p->stream_value(s);
			break;
		default:
			s << "?";
		}
	}
	void stream(ostream & s) const
	{
		PrettyPrintStreambuf buf(s);
		OpToken::stream(buf, this);
	}
private:
	OpToken(const OpToken &);
	OpToken & operator =(const OpToken &);
};

class TreeParser : public Parser
{
public:
	TreeParser()
	: Parser()
	, variables()
	, m_result()
	{
	}
	string result() const
	{
		stringstream ss;
		ss << m_result;
		return ss.str();
	}
protected:
	void assign(TokenPtr & lhs, TokenPtr & rhs)
	{
		variables[lhs->string_value()] = rhs;
	}
	void store_result(TokenPtr & p)
	{
		m_result = p;
	}
	TokenPtr add(TokenPtr &a, TokenPtr &b)
	{
		Token * ret = new OpToken('+', a, b);
		return ret;
	}
	TokenPtr sub(TokenPtr &a, TokenPtr &b)
	{
		Token * ret = new OpToken('-', a, b);
		return ret;
	}
	TokenPtr mult(TokenPtr &a, TokenPtr &b)
	{
		Token * ret = new OpToken('*', a, b);
		return ret;
	}
	TokenPtr div(TokenPtr &a, TokenPtr &b)
	{
		Token * ret = new OpToken('/', a, b);
		return ret;
	}
	TokenPtr fetch(TokenPtr &a)
	{
		const string & key = a->string_value();
		map<std::string, TokenPtr>::iterator i = variables.find(key);
		if(i == variables.end()) {
			stringstream ss;
			ss << "variable " << a->string_value() << " is not defined.";
			error(*a, ss.str());
		}
		return i->second; // This makes a value-based tree.
		//return a; // This makes a symbolic tree.
	}
	map<std::string, TokenPtr> variables;
	TokenPtr m_result;
};

} // namespace calc

int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	bool /*parse_debug = false, lex_debug = false,*/ tree = false;
	for(++argv; argv[0]; ++argv)
		/*if(*argv == std::string ("-p"))
			parse_debug = true;
		else if(*argv == std::string ("-s"))
			lex_debug = true;
		else*/ if(*argv == std::string ("-t"))
			tree = true;
		else try{
			if(!tree) {
				calc::IntParser parser;
				int ret = parser.parse(*argv);
				cout << parser.result() << endl;
				return ret;
			}else{
				calc::TreeParser parser;
				int ret = parser.parse(*argv);
				cout << parser.result() << endl;
				return ret;
			}
		}catch(const exception & e) {
			cerr << "Parsing failed : " << e.what() << endl;
		}
	return 0; // Succeed - it's just a test.
}
