#if !defined(calc_token_hpp_included)
#define calc_token_hpp_included

// Copyright (C) 2009  Philip Ashmore (contact@philipashmore.com)
// License: LGPLv3.  See LICENSE.txt for the full license.

#include <string>
#include <stdexcept>
#include <v3c/smart_ptr.hpp>
#include <v3c/span.hpp>

namespace calc {

using namespace v3c;

class Token;
typedef smart_ptr< Token, SingleThreadModelTraits > TokenPtr;

} // namespace calc

#include "calc_parser_tokens.hpp"

namespace calc {

using namespace std;
using namespace v3c;

class IntToken;
class StringToken;

class Token : public v3c::Span, public ParserTokens
{
	friend ostream & operator <<(ostream & s, const Token & t);

public:
	enum Type{ eLex, eInt, eString, eNext };

	/// Called from derived classes
	inline Token(Type t, int id_, const v3c::Position & p, unsigned int length)
	: v3c::Span(p, length)
	, m_type(t)
	, m_id(id_)
	{
	}
	/// Called from derived classes
	inline Token(int type_, int id_, const v3c::Span & s)
	: v3c::Span(s)
	, m_type(type_)
	, m_id(id_)
	{
	}
	/// Called directly
	inline Token(int id_, const v3c::Position & p, unsigned int length)
	: v3c::Span(p, length)
	, m_type(eLex)
	, m_id(id_)
	{
	}
	virtual ~Token() {}
	inline int type() const { return m_type; }
	virtual inline const char * name() const
	{
		switch(type()) {
		case eLex:
			return "lex";
		case eInt:
			return "int";
		case eString:
			return "string";
		default:
			return "<?>";
		}
	}
	inline int id() const { return m_id; }

	inline IntToken & asInt();
	inline StringToken & asString();
	inline const IntToken & asInt() const;
	inline const StringToken & asString() const;
	inline int & int_value();
	inline int int_value() const;
	inline string & string_value();
	inline const string & string_value() const;

	virtual void stream(ostream & s) const;
	virtual void stream_value(ostream & s) const;
protected:
	int m_type;
	int m_id;
};

inline ostream & operator <<(ostream & s, const Token & t)
{
	t.stream(s);
	return s;
}
inline ostream & operator <<(ostream & s, const TokenPtr & t)
{
	if(t)
		s << *t;
	else
		s << "<null>";
	return s;
}

class IntToken : public Token
{
public:
	IntToken(const v3c::Position & p, unsigned int length, int v)
	: Token(eInt, NUMBER, p, length)
	, value(v)
	{}
	IntToken(const v3c::Span & s, int v)
	: Token(eInt, NUMBER, s)
	, value(v)
	{}
	int value;
};

class StringToken : public Token
{
public:
	StringToken(const v3c::Position & p, unsigned int length, const string & s)
	: Token(eString, IDENTIFIER, p, length)
	, value(s)
	{}
	StringToken(const v3c::Position & p, const char * s, unsigned int len)
	: Token(eString, IDENTIFIER, p, len)
	, value(s, len)
	{}
	string value;
};

inline IntToken & Token::asInt()
{
	if(m_type == eInt)
		return static_cast< IntToken & >(*this);
	throw runtime_error(string("asInt called on ") + name());
}
inline StringToken & Token::asString()
{
	if(m_type == eString)
		return static_cast< StringToken & >(*this);
	throw runtime_error(string("asString called on ") + name());
}
inline const IntToken & Token::asInt() const
{
	if(m_type == eInt)
		return static_cast< const IntToken & >(*this);
	throw runtime_error(string("asInt called on ") + name());
}
inline const StringToken & Token::asString() const
{
	if(m_type == eString)
		return static_cast< const StringToken & >(*this);
	throw runtime_error(string("asString called on ") + name());
}
inline int & Token::int_value()
{
	return asInt().value;
}
inline string & Token::string_value()
{
	return asString().value;
}
inline int Token::int_value() const
{
	return asInt().value;
}
inline const string & Token::string_value() const
{
	return asString().value;
}

} // namespace calc

#endif // #if !defined(calc_token_hpp_included)
