// Copyright (C) 2009  Philip Ashmore (contact@philipashmore.com)
// License: LGPLv3.  See LICENSE.txt for the full license.

#include <v3c/trace_scope.hpp>
#include <iostream>
#include <v3c/i_chunk_stream.hpp>
#include <v3c/string_utils.hpp>

#include "bisonx_parser.hpp"

using namespace std;
using namespace v3c;

#include <v3c/cmdline/parser.hpp>

class CmdLineParser : public cmdline::Parser
{
public:
	CmdLineParser(int argc, char *argv[])
	: parser()
	, g()
	, input()
	{
		TraceWrap(__PRETTY_FUNCTION__);

		string args;
		argc--; // Skip program name.
		argv++;
		for(; argc; argc--, argv++) {
			if(strchr(argv[0], ' '))
				args += string("\"") + argv[0] + "\" ";
			else
				args += string(argv[0]) + " ";
		}
		string::size_type length = args.length();
		const char * start = args.c_str();
		i_chunk_stream i(start, start + length);
		switch_streams(&i, (ostream *)0);
		set_lex_debug(true);
		Parser::parse("dpso:r:", 0);
		if(!input.length())
			throw runtime_error("No input file!");
		trace << "Starting parse of \"" << input << "\"." << endl;
		int ret = parser.parse(g, input.c_str());
		if(ret)
			throw runtime_error(
				string_inline_printf("Parse failed with return code %d.", ret));
	}
	virtual bool on_opt(int c)
	{
		TraceWrap(__PRETTY_FUNCTION__);

		switch(c) {
		case 'd':
			trace_cerr << "Ignoring -d" << endl;
			break; // We always produce a header file - ignored.
		case 'p':
			parser.set_parse_debug(true);
			break;
		case 's':
			parser.set_lex_debug(true);
			break;
		default:
			return false;
		}
		return true;
	}
	virtual bool on_opt_arg(int c, const char * param)
	{
		TraceWrap(__PRETTY_FUNCTION__ << " : " << (char)c << " -> \"" << param << '"');

		switch(c) {
		case 'o': // TODO store where to write things.
			trace_cerr << "Ignoring -o " << param << endl;
			break;
		case 'r': // TODO report things.
			trace_cerr << "Ignoring -r " << param << endl;
			break;
		default:
			return false;
		}
		return true;
	}
	virtual bool on_arg(const char * param)
	{
		TraceWrap(__PRETTY_FUNCTION__ << " : \"" << param << '"');
		trace << "Will read \"" << param << "\"." << endl;
		input = param;
		return true;
	}
	bisonx::impl::Parser parser;
	bisonx::impl::Grammar g;
	string input;
};

int
main (int argc, char *argv[])
{
	CoutTraceWriter ctw("");

	TraceWrap(__PRETTY_FUNCTION__);

	try{
		CmdLineParser cmdline(argc, argv);
		if(!cmdline.input.length()) {
			trace_cerr << "Usage: bisonx [flags] <grammar file>" << endl;
			return 1;
		}
		trace_cout << "Parse succeeded!" << endl;
		trace_cout << cmdline.g << endl;
		return 0;
	}catch(const exception & e) {
		trace_cerr << "Parsing failed : " << e.what() << endl;
		return 2;
	}
	return 1;
}

namespace v3c {
namespace bisonx {
namespace impl {

void Parser::version_check(NodePtr & t)
{
	TraceWrap(__PRETTY_FUNCTION__);
	const string & ver = t->string_value();
	const char * version = ver.c_str();

	if (strverscmp(version, PACKAGE_VERSION) > 0) {
		trace_cerr << t->asSpan() << " : require bison(x) "
			<< version << ", but have " << PACKAGE_VERSION << endl;
    }
}
void Parser::symbol_make_alias(Token * p, NodePtr & q)
{
	//TraceWrap(__PRETTY_FUNCTION__);
	if(!p->is_symbol())
		fatal_at(q->asSpan(), "symbol_make_alias called with parameter 1: %s"
				, p->type_name());
	if(!q->is_string())
		fatal_at(q->asSpan(), "symbol_make_alias called with parameter 2: %s"
				, q->type_name());
	const string & name = q->string_value();
	Token * t = new Token(q->asSpan(), name);
	trace << '"' << name << "\" -> " << p->name() << endl;
	grammar->symbols[name] = t;
	t->user_token_number_set(Symbol::USER_NUMBER_ALIAS);
	t->m_pAlias = p;
}

/// Add another rule rhs to the current rule.
void Parser::rule_begin()
{
	TraceWrap(m_scanner << " : " << __PRETTY_FUNCTION__);
	if(!current_rule)
		fatal_at(v3c::Span(m_scanner, m_scanner), "rule_begin : no rule!");
	current_rule->productions.push_back(Production());
}
/*! Append a rule component to this rule

If the node is a string, convert it into a symbol ref, so we know where
undefined symbols are referenced.

Code nodes are inserted as is.

Anything else is an error.
*/
void Parser::rule_append(NodePtr & p)
{
	TraceWrap(__PRETTY_FUNCTION__);
	if(!current_rule)
		fatal_at(p->asSpan(), "rule_append : no rule!");
	switch(p->type()) {
	case Node::eCode:
		current_rule->productions.back().vec.push_back(p);
		break;
	case Node::eString:
		current_rule->productions.back().vec.push_back(new SymbolRef(p));
		break;
	case Node::eSymbolRef:
		current_rule->productions.back().vec.push_back(p);
		break;
	default:
		fatal_at(p->asSpan(), "rule_append called with %s"
				, p->type_name());
	}
}
void Parser::rule_prec_set(NodePtr & p)
{
	if(!current_rule)
		fatal_at(p->asSpan(), "rule_prec_set : no rule!");
	current_rule->productions.back().prec_set(p->int_value());
}
void Parser::rule_dprec_set(NodePtr & p)
{
	if(!current_rule)
		fatal_at(p->asSpan(), "rule_prec_set : no rule!");
	current_rule->productions.back().dprec_set(p->int_value());
}
void Parser::rule_merge_set(NodePtr & p)
{
	if(!current_rule)
		fatal_at(p->asSpan(), "rule_merge_set : no rule!");
	current_rule->productions.back().mergefunc = p->string_value();
}
void Parser::rule_end(NodePtr & p)
{
	(void)p;
	TraceWrap(__PRETTY_FUNCTION__);
}

ostream & operator <<(ostream & s, const Node & t)
{
	t.stream_span(s);
	s << " : " << t.type_name();
	switch(t.type()) {
	case Node::eLexToken:
		break;
	case Node::eToken:
	case Node::eRule:
	case Node::eSymbolRef:
		s << " [" << t.asSymbol().number() << "] " << t.asSymbol().name();
		break;
	case Node::eInt:
		s << " [" << t.id() << "] ";
		s << "( = " << t.int_value() << ')';
		break;
	case Node::eString:
		s << " [" << t.id() << "] ";
		s << '"' << t.string_value() << '"';
		break;
	case Node::eCode:
		s << " [" << t.id() << "] ";
		s << '"' << t.string_value() << '"';
		{
			const CodeToken & c = t.asCode();
			MarkerVec::const_iterator i = c.values.begin();
			MarkerVec::const_iterator j = c.values.end();
			for(; i != j; ++i)
				s << " [" << i->offset << "," << i->length << ']';
		}
		break;
	case Node::eTokenList:
		s << " : " << t.type_name();
		break;
	}
	return s;
}

} // namespace impl
} // namespace bisonx
} // namespace v3c
