#if !defined(v3c_bisonx_grammar_hpp_included)
#define v3c_bisonx_grammar_hpp_included

// Copyright (C) 2009  Philip Ashmore (contact@philipashmore.com)
// License: LGPLv3.  See LICENSE.txt for the full license.

#include "bisonx_token.hpp"
#include <map>
#include <v3c/pretty_print_streambuf.hpp>

namespace v3c {
namespace bisonx {
namespace impl {

using namespace std;
using namespace v3c;

// SymbolClass was a way of noting whether a rule(non-terminal) or a
// token(terminal) is to be created.
enum SymbolClass
{ unknown_sym
, token_sym   /* terminal symbol */
, nterm_sym   /* non-terminal (rule) */
};

enum Report
{
report_none             = 0,
report_states           = 1 << 0,
report_itemsets         = 1 << 1,
report_look_ahead_tokens= 1 << 2,
report_solved_conflicts = 1 << 3,
report_all              = ~0
};

enum Trace
{
trace_none      = 0,
trace_scan      = 1 << 0,
trace_parse     = 1 << 1,
trace_resource  = 1 << 2,
trace_sets      = 1 << 3,
trace_bitsets   = 1 << 4,
trace_tools     = 1 << 5,
trace_automaton = 1 << 6,
trace_grammar   = 1 << 7,
trace_time      = 1 << 8,
trace_skeleton  = 1 << 9,
trace_m4        = 1 << 10,
trace_all       = ~0
};

class Production;  // A sequence of rule rhs components with attributes
class Rule;        // A rule containing a RuleRhs
class Grammar;     // The whole shebang.

class Symbol : public Node
{
protected:
	inline Symbol(NodePtr & p, Node::Type t);
	inline Symbol(const v3c::Span & span, const string & name_, Node::Type t);

public:
	inline const string & name() const;

	inline void assoc_set(Assoc a);
	inline Assoc assoc() const;

	inline void prec_set(int p);
	inline int prec() const;

	inline void number_set(unsigned int n);
	inline unsigned int number() const;
	enum
	/* Undefined user number.  */
	{ USER_NUMBER_UNDEFINED = -1

	/* `symbol->user_token_number == USER_NUMBER_ALIAS' means this symbol
	*has* (not is) a string literal alias.  For instance, `%token foo
	"foo"' has `"foo"' numbered regularly, and `foo' numbered as
	USER_NUMBER_ALIAS.  */
	, USER_NUMBER_ALIAS = -9991

	/* Undefined internal token number.  */
	, NUMBER_UNDEFINED = -1
	};
protected:
	Assoc m_assoc;
	int m_prec;
	int m_number;
	string m_name;
};
PrettyPrintStreambuf & operator <<(PrettyPrintStreambuf & s, const Symbol & t);

typedef smart_ptr< Symbol, SingleThreadModelTraits > SymbolPtr;
typedef vector< SymbolPtr > SymbolPtrVec;
PrettyPrintStreambuf & operator <<(PrettyPrintStreambuf & s, const SymbolPtr & g);

typedef map< string, SymbolPtr > SymbolMap;   // Tokens + rules keyed by name.

class Token : public Symbol
{
public:
	inline Token(NodePtr & p);
	inline Token(const v3c::Span & span, const string & name_);
	inline void user_token_number_set(int i);
	inline int user_token_number() const;

	/* Points to the other in the identifier-symbol pair for an alias.
		Special value USER_NUMBER_ALIAS in the identifier half of the
		identifier-symbol pair for an alias.  */
	Token *m_pAlias;
protected:
	int m_user_token_number; // May be USER_NUMBER_ALIAS
private:
	Token(const Token &);
	Token & operator =(const Token &);
};

class TokenList : public Node
{
public:
	inline TokenList();
	inline TokenList(SymbolPtr & p);

	inline void assoc_set(Assoc a);
	inline Assoc assoc() const;

	inline void prec_set(int p);
	inline int prec() const;
	SymbolPtrVec tokens;
protected:
	Assoc m_assoc;
	int m_prec;
};

/// A sequence of rule rhs components with attributes
class Production
{
public:
	inline Production();
	inline void assoc_set(Assoc a);
	inline Assoc assoc() const;
	inline void prec_set(int p);
	inline int prec() const;
	inline void dprec_set(int p);
	inline int dprec() const;
	NodePtrVec vec;
	string mergefunc;
	Assoc m_assoc;
	int m_prec, m_dprec;
};
PrettyPrintStreambuf & operator <<(PrettyPrintStreambuf & s, const Production & g);

typedef vector< Production > ProductionVec;
PrettyPrintStreambuf & operator <<(PrettyPrintStreambuf & s, const ProductionVec & g);

class Rule : public Symbol
{
public:
	inline Rule(NodePtr & r);
	ProductionVec productions;
};
PrettyPrintStreambuf & operator <<(PrettyPrintStreambuf & s, const Rule & g);

class SymbolRef : public Symbol
{
public:
	inline SymbolRef(NodePtr & p);
	inline void used_set(bool b = true);
	inline bool used() const;
protected:
	bool m_bUsed;
};

typedef smart_ptr< CodeToken, SingleThreadModelTraits > CodeTokenPtr;
typedef vector< CodeTokenPtr > CodeTokenPtrVec;

typedef smart_ptr< StringToken, SingleThreadModelTraits > StringTokenPtr;
ostream & operator <<(ostream & s, const StringTokenPtr & t);

typedef map< string, string > DefinesMap;
ostream & operator <<(ostream & s, const DefinesMap & g);

class Grammar
{
public:
	inline Grammar();
	inline void reset();

	StringTokenPtr prologue, epilogue;
	SymbolMap symbols;
	DefinesMap defines;
	string start_symbol;
	bool default_prec;
	CodeTokenPtrVec initial_actions, actions;
};

ostream & operator <<(ostream & s, const Grammar & g);

} // namespace impl
} // namespace bisonx
} // namespace v3c

#include "bisonx_grammar.impl.hpp"

#endif // #if !defined(v3c_bisonx_grammar_hpp_included)
