#if !defined(v3c_bisonx_grammar_impl_hpp_included)
#define v3c_bisonx_grammar_impl_hpp_included

// Copyright (C) 2009  Philip Ashmore (contact@philipashmore.com)
// License: LGPLv3.  See LICENSE.txt for the full license.

#if !defined(v3c_bisonx_grammar_hpp_included)
#error This file may only be included from bisonx_grammar.hpp
#endif

namespace v3c {
namespace bisonx {
namespace impl {

////////////////////////////////////////////////////////////////////////////////
// Node
////////////////////////////////////////////////////////////////////////////////
inline Token & Node::asToken()
{
	if(m_type == eToken)
		return static_cast< Token & >(*this);
	throw runtime_error(string("asToken called on ") + type_name());
}
inline const Token & Node::asToken() const
{
	if(m_type == eToken)
		return static_cast< const Token & >(*this);
	throw runtime_error(string("asToken called on ") + type_name());
}

inline TokenList & Node::asTokenList()
{
	if(m_type == eTokenList)
		return static_cast< TokenList & >(*this);
	throw runtime_error(string("asTokenList called on ") + type_name());
}
inline const TokenList & Node::asTokenList() const
{
	if(m_type == eTokenList)
		return static_cast< const TokenList & >(*this);
	throw runtime_error(string("asTokenList called on ") + type_name());
}

inline Rule & Node::asRule()
{
	if(m_type == eRule)
		return static_cast< Rule & >(*this);
	throw runtime_error(string("asRule called on ") + type_name());
}
inline const Rule & Node::asRule() const
{
	if(m_type == eRule)
		return static_cast< const Rule & >(*this);
	throw runtime_error(string("asRule called on ") + type_name());
}

inline SymbolRef & Node::asSymbolRef()
{
	if(m_type == eSymbolRef)
		return static_cast< SymbolRef & >(*this);
	throw runtime_error(string("asSymbolRef called on ") + type_name());
}
inline const SymbolRef & Node::asSymbolRef() const
{
	if(m_type == eSymbolRef)
		return static_cast< const SymbolRef & >(*this);
	throw runtime_error(string("asSymbolRef called on ") + type_name());
}

inline Symbol & Node::asSymbol()
{
	if(m_type == eToken || m_type == eRule || m_type == eSymbolRef)
		return static_cast< Symbol & >(*this);
	throw runtime_error(string("asSymbol called on ") + type_name());
}
inline const Symbol & Node::asSymbol() const
{
	if(m_type == eToken || m_type == eRule || m_type == eSymbolRef)
		return static_cast< const Symbol & >(*this);
	throw runtime_error(string("asSymbol called on ") + type_name());
}

inline string Node::string_value() const
{
	if(m_type == eString || m_type == eCode)
		return asString().value;
	return asSymbolRef().name();
}

////////////////////////////////////////////////////////////////////////////////
// Symbol
////////////////////////////////////////////////////////////////////////////////
inline Symbol::Symbol(NodePtr & p, Node::Type t)
: Node(p->asSpan(), t)
, m_assoc(undef_assoc)
, m_prec(0)
, m_number(0)
, m_name(p->string_value())
{
}
inline Symbol::Symbol(const v3c::Span & span, const string & name_, Node::Type t)
: Node(span, t)
, m_assoc(undef_assoc)
, m_prec(0)
, m_number(0)
, m_name(name_)
{
}
inline const string & Symbol::name() const { return m_name; }

inline void Symbol::assoc_set(Assoc a) { m_assoc = a; }
inline Assoc Symbol::assoc() const { return m_assoc; }

inline void Symbol::prec_set(int p) { m_prec = p; }
inline int Symbol::prec() const { return m_prec; }

inline void Symbol::number_set(unsigned int n) { m_number = n; }
inline unsigned int Symbol::number() const { return m_number; }

////////////////////////////////////////////////////////////////////////////////
// Token
////////////////////////////////////////////////////////////////////////////////
inline Token::Token(NodePtr & p)
: Symbol(p, eToken)
, m_pAlias()
, m_user_token_number(0)
{
}
inline Token::Token(const v3c::Span & span, const string & name_)
: Symbol(span, name_, eToken)
, m_pAlias()
, m_user_token_number(0)
{
}
inline void Token::user_token_number_set(int i) { m_user_token_number = i; }
inline int Token::user_token_number() const { return m_user_token_number; }

////////////////////////////////////////////////////////////////////////////////
// TokenList
////////////////////////////////////////////////////////////////////////////////
inline TokenList::TokenList()
: Node(eTokenList)
, tokens()
, m_assoc(undef_assoc)
, m_prec(0)
{
}
inline TokenList::TokenList(SymbolPtr & p)
: Node(p->asSpan(), eTokenList)
, tokens()
, m_assoc(undef_assoc)
, m_prec(0)
{
	tokens.push_back(p);
}

inline void TokenList::assoc_set(Assoc a)
{
	m_assoc = a;
	SymbolPtrVec::iterator i = tokens.begin();
	SymbolPtrVec::iterator j = tokens.end();
	for(; i != j; ++i)
		(*i)->assoc_set(a);
}
inline Assoc TokenList::assoc() const { return m_assoc; }

inline void TokenList::prec_set(int p) { m_prec = p; }
inline int TokenList::prec() const { return m_prec; }

////////////////////////////////////////////////////////////////////////////////
// Production
////////////////////////////////////////////////////////////////////////////////
inline Production::Production()
: vec()
, mergefunc()
, m_assoc(undef_assoc)
, m_prec()
, m_dprec()
{
}
inline void Production::assoc_set(Assoc a) { m_assoc = a; }
inline Assoc Production::assoc() const { return m_assoc; }

inline void Production::prec_set(int p) { m_prec = p; }
inline int Production::prec() const { return m_prec; }

inline void Production::dprec_set(int p) { m_dprec = p; }
inline int Production::dprec() const { return m_dprec; }

////////////////////////////////////////////////////////////////////////////////
// Rule
////////////////////////////////////////////////////////////////////////////////
inline Rule::Rule(NodePtr & r)
: Symbol(r, eRule)
, productions()
{
}

////////////////////////////////////////////////////////////////////////////////
// SymbolRef
////////////////////////////////////////////////////////////////////////////////
inline SymbolRef::SymbolRef(NodePtr & p)
: Symbol(p, eSymbolRef)
, m_bUsed(false)
{
}
inline void SymbolRef::used_set(bool b) { m_bUsed = b; }
inline bool SymbolRef::used() const { return m_bUsed; }

////////////////////////////////////////////////////////////////////////////////
// Grammar
////////////////////////////////////////////////////////////////////////////////
inline Grammar::Grammar()
: prologue()
, epilogue()
, symbols()
, defines()
, start_symbol()
, default_prec()
, initial_actions()
, actions()
{
}
inline void Grammar::reset()
{
	*this = Grammar();
}

} // namespace impl
} // namespace bisonx
} // namespace v3c

#endif // #if !defined(v3c_bisonx_grammar_impl_hpp_included)
