#if !defined(v3c_bisonx_scanner_hpp_included)
#define v3c_bisonx_scanner_hpp_included

// Copyright (C) 2009  Philip Ashmore (contact@philipashmore.com)
// License: LGPLv3.  See LICENSE.txt for the full license.

#include <v3c/flexx/flexx.hpp>

#include <stdexcept>
#include <sstream>
#include <climits>
#include "bisonx_token.hpp"
#include "v3c_bisonx_impl_parser_tokens.hpp"

namespace v3c {
namespace bisonx {
namespace impl {

using namespace std;

void fatal_at(const v3c::Span & loc, const char * format, ...)
	__attribute__ ((__format__ (__printf__, 2, 3)));

// TODO i18n
inline size_t mbsnwidth(const char *buf, size_t bufsize, int flags)
{
	(void)buf;
	(void)flags;
	return bufsize;
}

class Parser;

class Scanner
: public v3c::Flexx
, public ParserTokens
{
	FLEXX_IMPL(NodePtr);
public:
	/* By default yylex returns int, we use m_token_type.
	   Unfortunately yyterminate by default returns 0, which is
	   not of m_token_type. */
	enum { eTerminateValue = END };
	Scanner(Parser & parser);
	virtual void error(const char* msg)
	{
		LexerError("%s", msg); // Maybe replace LexerError() with error()?
	}
	virtual void error_char(const char* msg)
	{
		stringstream ss;
		ss << "At " << *this << " : " << msg << ' ';
		if(isprint(*yytext))
			ss << '\'' << *yytext << '\'';
		else
			ss << "0x" << hex << (unsigned int)(*yytext);
		ss << '.';
		throw runtime_error(ss.str());
	}
	int spaces_per_tab() const { return m_spaces_per_tab; }
	void set_spaces_per_tab(int n) { m_spaces_per_tab = n; }
protected:
// TODO i18n
	static inline string _(const string & X) { return X; }
	static inline string N_(const string & X) { return X; }
	static inline string quote(const string & X) { return X; }
	static inline const char * _(const char * X) { return X; }
	static inline const char * N_(const char * X) { return X; }
	static inline const char * quote(const char * X) { return X; }
	unsigned int mbsnwidth(int /*FLags*/) { return YYLeng(); }
	static inline int add_column_width(int column, char const *buf, size_t bufsize);

	void warn_at(const v3c::Span & loc, const char * format, ...)
		__attribute__ ((__format__ (__printf__, 3, 4)));
	void complain_at(const v3c::Span & loc, const char * format, ...)
		__attribute__ ((__format__ (__printf__, 3, 4)));

	inline void note_rule_start()
	{
		m_rule_length = 0;
		m_rule_start = *this;
	}
	/* Increment the rule length by one, checking for overflow.  */
	inline void
	increment_rule_length()
	{
		m_rule_length++;
	
		/* Don't allow m_rule_length == INT_MAX, since that might cause
			confusion with strtol if INT_MAX == LONG_MAX.  */
		if (m_rule_length == INT_MAX)
			rule_length_overflow();
	}
	void rule_length_overflow();
	inline bool handle_action_dollar();
	void handle_dollar();
	bool handle_action_at();
	void handle_at();
	void handle_syncline(char *);
	NodePtr scan_integer(int base, int token);
	int convert_ucn_to_byte(const char * hex_text);
	void unexpected_end(const v3c::Position & start, const char * msgid, const char * token_end);
	void unexpected_eof(const v3c::Position & start, const char *);
	void unexpected_newline(const v3c::Position & start, const char *);

	//inline LexTokenPtr &
	//symbol_get(const v3c::Position & loc, const string & key);
	NodePtr make_token(int t);
	NodePtr make_token(const v3c::Position & loc, int id);
	NodePtr make_string_token(const v3c::Position & loc, const string & text, int token);
	NodePtr make_int_token(const v3c::Span & loc, int n, int token);
	//void update_position(const char * text, unsigned int length);
	//void newline();

	inline string token_name(int type) const;

	Parser & m_parser;

	/* Within well-formed rules, RULE_LENGTH is the number of values in
	the current rule so far, which says where to find `$0' with respect
	to the top of the stack.  It is not the same as the rule->length in
	the case of mid rule actions.
	Outside of well-formed rules, RULE_LENGTH has an undefined value.  */
	unsigned int m_rule_length;

	/* Nesting level of the current code in braces.  */
	int m_braces_level;

	/* Parent context state, when applicable.  */
	int m_context_state;

	/* Token type to return, when applicable.  */
	int m_token_type;

	/* Location of most recent identifier, when applicable.  */
	v3c::Position m_id_loc;

	/* Where containing code started, when applicable.  Its initial
		value is relevant only when yylex is invoked in the SC_EPILOGUE
		start condition.  */
	v3c::Position m_code_start;

	/* Where containing comment or string or character literal started,
		when applicable.  */
	v3c::Position m_token_start;

	bool m_bWarningIssued, m_bComplaintIssued;

	/* Keeps track of the maximum number of semantic values to the left of
	a handle (those referenced by $0, $-1, etc.) are required by the
	semantic actions of this grammar. */
	int max_left_semantic_context;


	/* The location of the most recently saved token, if it was a
	BRACED_CODE token; otherwise, this has an unspecified value.  */
	v3c::Position last_braced_code_loc;

	string last_string, cur_string;

	NodePtr val;

	v3c::Position m_rule_start;
};

} // namespace impl
} // namespace bisonx
} // namespace v3c

#endif // #if !defined(v3c_bisonx_scanner_hpp_included)
