%{                                            /* -*- C++ -*- */
#include <cstdlib>
#include <errno.h>
#include <limits.h>
#include <string>
#define IN_BISONX_SCANNER
#include "bisonx_parser.hpp"

#define verify(cond) extern char condition_is_false[-1 + (cond)]; \
(void)condition_is_false

#define STRING_GROW   \
	cur_string += string(yytext, yyleng)
//	trace << "STRING_GROW : cur_string [" << cur_string << "]." << endl

#define STRING_FINISH \
do { \
	last_string = cur_string; \
	cur_string = ""; \
	trace << "STRING_FINISH : last_string [" << last_string << "]." << endl; \
} while (0)

#define STRING_FREE \
	last_string = ""

namespace v3c {
namespace bisonx {
namespace impl {

inline string
Scanner::token_name(int type) const
{
	return m_parser.token_name(type);
}
map<int, string > lexmap;

} // namespace impl
} // namespace bisonx
} // namespace v3c

%}

%option debug nodefault nounput noyywrap never-interactive
%option outfile="bisonx_scanner.cpp"
%option c++ yyclass="v3c::bisonx::impl::Scanner"

%x SC_COMMENT SC_LINE_COMMENT SC_YACC_COMMENT
%x SC_STRING SC_CHARACTER
%x SC_AFTER_IDENTIFIER
%x SC_ESCAPED_STRING SC_ESCAPED_CHARACTER
%x SC_PRE_CODE SC_BRACED_CODE SC_PROLOGUE SC_EPILOGUE

letter	  [.abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_]
id	  {letter}({letter}|[0-9])*
directive %{letter}({letter}|[0-9]|-)*
int	  [0-9]+

/* POSIX says that a tag must be both an id and a C union member, but
   historically almost any character is allowed in a tag.  We disallow
   NUL and newline, as this simplifies our implementation.  */
tag	 [^\0\n>]+

/* Zero or more instances of backslash-newline.  Following GCC, allow
   white space between the backslash and the newline.  */
splice	 (\\[ \f\t\v]*\n)*

%%
  /*-----------------------.
  | Scanning white space.  |
  `-----------------------*/

<INITIAL,SC_AFTER_IDENTIFIER,SC_PRE_CODE>
{
  /* Comments and white space.  */
  ","	       {
	warn_at(v3c::Span(*this, yyleng), "%s", _("stray `,' treated as white space"));
}
  [ \f\n\t\v]  |
  "//".*       ;
  "/*" {
	m_token_start = *this;
	m_context_state = YY_START;
	BEGIN(SC_YACC_COMMENT);
  }

  /* #line directives are not documented, and may be withdrawn or
     modified in future versions of Bison.  */
  ^"#line "{int}" \"".*"\"\n" {
    handle_syncline (yytext + sizeof "#line " - 1);
  }
}


  /*----------------------------.
  | Scanning Bison directives.  |
  `----------------------------*/
<INITIAL>
{
  "%binary"               return make_token(PERCENT_NONASSOC);
  "%debug"                return make_token(PERCENT_DEBUG);
  "%default"[-_]"prec"    return make_token(PERCENT_DEFAULT_PREC);
  "%define"               return make_token(PERCENT_DEFINE);
  "%defines"              return make_token(PERCENT_DEFINES);
  "%destructor"          {
	warn_at(v3c::Span(*this, yyleng), "Ignoring \"%%destructor\".");
	m_token_type = PERCENT_DESTRUCTOR;
	BEGIN(SC_PRE_CODE);
}
  "%dprec"                return make_token(PERCENT_DPREC);
  "%error"[-_]"verbose"   return make_token(PERCENT_ERROR_VERBOSE);
  "%expect"               return make_token(PERCENT_EXPECT);
  "%expect"[-_]"rr"       return make_token(PERCENT_EXPECT_RR);
  "%file-prefix"          return make_token(PERCENT_FILE_PREFIX);
  "%fixed"[-_]"output"[-_]"files"  return make_token(PERCENT_YACC);
  "%initial-action"       {
	m_token_type = PERCENT_INITIAL_ACTION;
	BEGIN(SC_PRE_CODE);
}
  "%glr-parser"           {
	warn_at(v3c::Span(*this, yyleng), "Ignoring \"%%glr-parser\".");
	//return PERCENT_GLR_PARSER;
}
  "%left"                 return make_token(PERCENT_LEFT);
  "%lex-param"            {
	m_token_type = PERCENT_LEX_PARAM;
	BEGIN(SC_PRE_CODE);
}
  "%locations"            return make_token(PERCENT_LOCATIONS);
  "%merge"                return make_token(PERCENT_MERGE);
  "%name"[-_]"prefix"     return make_token(PERCENT_NAME_PREFIX);
  "%no"[-_]"default"[-_]"prec"	return make_token(PERCENT_NO_DEFAULT_PREC);
  "%no"[-_]"lines"        return make_token(PERCENT_NO_LINES);
  "%nonassoc"             return make_token(PERCENT_NONASSOC);
  "%nondeterministic-parser"   {
	return make_token(PERCENT_NONDETERMINISTIC_PARSER);
}
  "%nterm"                return make_token(PERCENT_NTERM);
  "%output"               return make_token(PERCENT_OUTPUT);
  "%parse-param"          {
	m_token_type = PERCENT_PARSE_PARAM;
	BEGIN(SC_PRE_CODE);
}
  "%prec"                 {
	m_rule_length--;
	return make_token(PERCENT_PREC);
}
  "%printer"              {
	warn_at(v3c::Span(*this, yyleng), "Ignoring \"%%printer\".");
	m_token_type = PERCENT_PRINTER;
	BEGIN(SC_PRE_CODE);
}
  "%pure"[-_]"parser"     return make_token(PERCENT_PURE_PARSER);
  "%require"              return make_token(PERCENT_REQUIRE);
  "%right"                return make_token(PERCENT_RIGHT);
  "%skeleton"             return make_token(PERCENT_SKELETON);
  "%start"                return make_token(PERCENT_START);
  "%term"                 return make_token(PERCENT_TOKEN);
  "%token"                return make_token(PERCENT_TOKEN);
  "%token"[-_]"table"     return make_token(PERCENT_TOKEN_TABLE);
  "%type"                 {
	warn_at(v3c::Span(*this, yyleng), "Ignoring \"%%type\".");
	//return PERCENT_TYPE;
}
  "%union"                {
	warn_at(v3c::Span(*this, yyleng), "Ignoring \"%%type\".");
	m_token_type = PERCENT_UNION;
	BEGIN(SC_PRE_CODE);
}
  "%verbose"              return make_token(PERCENT_VERBOSE);
  "%yacc"                 return make_token(PERCENT_YACC);

  {directive} {
    complain_at(v3c::Span(*this, yyleng), _("invalid directive: %s"), quote (yytext));
  }

  "="                     return make_token(EQUAL);
  "|"                     note_rule_start(); return make_token(PIPE);
  ";"                     return make_token(SEMICOLON);

  {id} {
    make_string_token(*this, string(yytext, yyleng), ID);
    increment_rule_length();
    BEGIN(SC_AFTER_IDENTIFIER);
  }

  {int} {
    return scan_integer(10, INT);
  }
  0[xX][0-9abcdefABCDEF]+ {
    return scan_integer(16, INT);
  }

  /* Characters.  We don't check there is only one.  */
  "'"	      {
	STRING_GROW;
	m_token_start = *this;
	BEGIN(SC_ESCAPED_CHARACTER);
}
  /* Strings. */
  "\""	      {
	m_token_start = *this;
	BEGIN(SC_ESCAPED_STRING);
}
  /* Prologue. */
  "%{"        {
	m_code_start = *this;
	BEGIN(SC_PROLOGUE);
}
  /* Code in between braces.  */
  "{" {
	STRING_GROW;
	m_token_type = BRACED_CODE;
	m_braces_level = 0;
	m_code_start = *this;
	val = new CodeToken(*this, yytext, 0, BRACED_CODE);
	BEGIN(SC_BRACED_CODE);
	trace << "SC_BRACED_CODE : " << val << endl;
  }

  /* A type. */
  "<"{tag}">" {
    //cur_string += string(yytext + 1, yyleng - 2);
    STRING_FINISH;
    //val->uniqstr = uniqstr_new (last_string);
    //STRING_FREE;
    //return TYPE;
  }

  "%%" {
    static int percent_percent_count;
    if (++percent_percent_count == 2)
      BEGIN(SC_EPILOGUE);
    return make_token(PERCENT_PERCENT);
  }

  . {
    complain_at(v3c::Span(*this, yyleng), _("invalid character: %s"), quote (yytext));
  }

  <<EOF>> {
	trace << "Reached EOF in " << lexmap[YY_START] << " (" << YY_START << ")." << endl;
    terminate();
  }
}


  /*-----------------------------------------------------------------.
  | Scanning after an identifier, checking whether a colon is next.  |
  `-----------------------------------------------------------------*/

<SC_AFTER_IDENTIFIER>
{
  ":" {
    note_rule_start();
    BEGIN(INITIAL);
    return make_token(ID_COLON);
  }
  . {
	//scanner_cursor.column -= mbsnwidth(yytext, yyleng, 0);
	yyless (0);
	//column--;
	BEGIN(INITIAL);
	return make_token(ID);
  }
  <<EOF>> {
	BEGIN(INITIAL);
	return make_token(ID);
  }
}


  /*---------------------------------------------------------------.
  | Scanning a Yacc comment.  The initial `/ *' is already eaten.  |
  `---------------------------------------------------------------*/

<SC_YACC_COMMENT>
{
  "*/"     BEGIN(m_context_state);
  .|\n	   ;
  <<EOF>>  unexpected_eof(m_token_start, "*/"); BEGIN(m_context_state);
}


  /*------------------------------------------------------------.
  | Scanning a C comment.  The initial `/ *' is already eaten.  |
  `------------------------------------------------------------*/

<SC_COMMENT>
{
  "*"{splice}"/"  STRING_GROW; BEGIN(m_context_state);
  <<EOF>>	  unexpected_eof(m_token_start, "*/"); BEGIN(m_context_state);
}


  /*--------------------------------------------------------------.
  | Scanning a line comment.  The initial `//' is already eaten.  |
  `--------------------------------------------------------------*/

<SC_LINE_COMMENT>
{
  "\n"		 STRING_GROW; BEGIN(m_context_state);
  {splice}	 STRING_GROW;
  <<EOF>>	 BEGIN(m_context_state);
}


  /*------------------------------------------------.
  | Scanning a Bison string, including its escapes. |
  | The initial quote is already eaten.             |
  `------------------------------------------------*/

<SC_ESCAPED_STRING>
{
  "\"" {
	STRING_FINISH;
	trace << "Finished string : [" << last_string << "]." << endl;
	increment_rule_length();
	BEGIN(INITIAL);
	return make_string_token(m_token_start, last_string, STRING);
  }
  .	{
	STRING_GROW;
}
  \n		unexpected_newline(m_token_start, "\"");	BEGIN(INITIAL);
  <<EOF>>	unexpected_eof(m_token_start, "\"");	BEGIN(INITIAL);
}

  /*----------------------------------------------------------.
  | Scanning a Bison character literal, decoding its escapes. |
  | The initial quote is already eaten.                       |
  `----------------------------------------------------------*/
<SC_ESCAPED_CHARACTER>
{
  "'" {
	//unsigned char last_string_1;
	STRING_GROW;
	STRING_FINISH;
	//val = symbol_get(m_token_start, last_string);
	//class_set(m_token_start, val, token_sym, false);
	//last_string_1 = last_string[1];
	// Convention, it seems, sets the user token number of the character literal
	// to it's character value.
	//val->asSymbol().user_token_number_set(last_string_1);
	increment_rule_length();
	BEGIN(INITIAL);
	trace << *this << " : end of char-spec." << endl;
    return make_string_token(*this, last_string, ID);
  }
  \n		unexpected_newline (m_token_start, "'");	BEGIN(INITIAL);
  <<EOF>>	unexpected_eof(m_token_start, "'");	BEGIN(INITIAL);
}

<SC_ESCAPED_CHARACTER,SC_ESCAPED_STRING>
{
  \0	    complain_at(v3c::Span(*this, yyleng), "%s", _("invalid null character"));
}


  /*----------------------------.
  | Decode escaped characters.  |
  `----------------------------*/

<SC_ESCAPED_STRING,SC_ESCAPED_CHARACTER>
{
  \\[0-7]{1,3} {
    unsigned long int c = strtoul (yytext + 1, NULL, 8);
    if (UCHAR_MAX < c)
      complain_at(v3c::Span(*this, yyleng), _("invalid escape sequence: %s"), quote (yytext));
    else if (! c)
      complain_at(v3c::Span(*this, yyleng), _("invalid null character: %s"), quote (yytext));
    else
      cur_string += (char)c;
  }

  \\x[0-9abcdefABCDEF]+ {
    verify (UCHAR_MAX < ULONG_MAX);
    unsigned long int c = strtoul (yytext + 2, NULL, 16);
    if (UCHAR_MAX < c)
      complain_at(v3c::Span(*this, yyleng), _("invalid escape sequence: %s"), quote (yytext));
    else if (! c)
      complain_at(v3c::Span(*this, yyleng), _("invalid null character: %s"), quote (yytext));
    else
      cur_string += (char)c;
  }

  \\a	cur_string += '\a';
  \\b	cur_string += '\b';
  \\f	cur_string += '\f';
  \\n	cur_string += '\n';
  \\r	cur_string += '\r';
  \\t	cur_string += '\t';
  \\v	cur_string += '\v';

  /* \\[\"\'?\\] would be shorter, but it confuses xgettext.  */
  \\("\""|"'"|"?"|"\\")  cur_string += yytext[1];

  \\(u|U[0-9abcdefABCDEF]{4})[0-9abcdefABCDEF]{4} {
    int c = convert_ucn_to_byte (yytext);
    if (c < 0)
      complain_at(v3c::Span(*this, yyleng), _("invalid escape sequence: %s"), quote (yytext));
    else if (! c)
      complain_at(v3c::Span(*this, yyleng), _("invalid null character: %s"), quote (yytext));
    else
      cur_string += c;
  }
  \\(.|\n)	{
    complain_at(v3c::Span(*this, yyleng), _("unrecognized escape sequence: %s"), quote (yytext));
    STRING_GROW;
  }
}

  /*--------------------------------------------.
  | Scanning user-code characters and strings.  |
  `--------------------------------------------*/

<SC_CHARACTER,SC_STRING>
{
  {splice}|\\{splice}[^\n$@\[\]]	{
	STRING_GROW;
	}
}

<SC_CHARACTER>
{
  "'"			STRING_GROW; BEGIN(m_context_state);
  \n		unexpected_newline(m_token_start, "'"); BEGIN(m_context_state);
  <<EOF>>	unexpected_eof(m_token_start, "'"); BEGIN(m_context_state);
}

<SC_STRING>
{
  "\""		STRING_GROW; BEGIN(m_context_state);
  \n		unexpected_newline (m_token_start, "\""); BEGIN(m_context_state);
  .			STRING_GROW;
  <<EOF>>	unexpected_eof(m_token_start, "\""); BEGIN(m_context_state);
}


  /*---------------------------------------------------.
  | Strings, comments etc. can be found in user code.  |
  `---------------------------------------------------*/

<SC_BRACED_CODE,SC_PROLOGUE,SC_EPILOGUE>
{
  "'" {
    STRING_GROW;
    m_context_state = YY_START;
    m_token_start = *this;
    BEGIN(SC_CHARACTER);
  }
  "\"" {
    STRING_GROW;
    m_context_state = YY_START;
    m_token_start = *this;
    BEGIN(SC_STRING);
  }
  "/"{splice}"*" {
    STRING_GROW;
    m_context_state = YY_START;
    m_token_start = *this;
    BEGIN(SC_COMMENT);
  }
  "/"{splice}"/" {
    STRING_GROW;
    m_context_state = YY_START;
    BEGIN(SC_LINE_COMMENT);
  }
}


  /*---------------------------------------------------------------.
  | Scanning after %union etc., possibly followed by white space.  |
  | For %union only, allow arbitrary C code to appear before the   |
  | following brace, as an extension to POSIX.			   |
  `---------------------------------------------------------------*/

<SC_PRE_CODE>
{
  . {
	bool valid = yytext[0] == '{' || m_token_type == PERCENT_UNION;
	//scanner_cursor.column -= mbsnwidth (yytext, yyleng, 0);
	yyless (0);
	//column--;
	if (valid) {
		m_braces_level = -1;
		m_code_start = *this;
		val = new CodeToken(*this, yytext, yyleng, BRACED_CODE);
		BEGIN(SC_BRACED_CODE);
		trace << "SC_BRACED_CODE : " << val << endl;
	}else{
		complain_at(v3c::Span(*this, yyleng), _("missing `{' in %s"),
				token_name(m_token_type).c_str());
		cur_string += "{}";
		STRING_FINISH;
		BEGIN(INITIAL);
		return make_string_token(*this, last_string, m_token_type);
	}
}

  <<EOF>>  unexpected_eof(*this, "{}"); BEGIN(INITIAL);
}


  /*---------------------------------------------------------------.
  | Scanning some code in braces (%union and actions). The initial |
  | "{" is already eaten.                                          |
  `---------------------------------------------------------------*/

<SC_BRACED_CODE>
{
  "{"|"<"{splice}"%"  STRING_GROW; m_braces_level++;
  "%"{splice}">"      STRING_GROW; m_braces_level--;
  "}" {
    bool outer_brace = --m_braces_level < 0;

    cur_string += '}';

	if (outer_brace) {
		STRING_FINISH;
		increment_rule_length();
		BEGIN(INITIAL);
		CodeToken & c = val->asCode();
		c.value += last_string;
		c.end = *this;
		NodePtr ret = val;
		val = 0;
		return ret;
	}
}

  /* Tokenize `<<%' correctly (as `<<' `%') rather than incorrectly
     (as `<' `<%').  */
  "<"{splice}"<"  STRING_GROW;

  "$"("<"{tag}">")?(-?[0-9]+|"$")  {
		trace << ">> " << *this << " : " << val << endl;
		handle_dollar();
	}
  "@"(-?[0-9]+|"$")		   handle_at();

  "$"  {
    warn_at(v3c::Span(*this, yyleng), "%s", _("stray `$'"));
    cur_string += "$][";
  }
  "@"  {
    warn_at(v3c::Span(*this, yyleng), "%s", _("stray `@'"));
    cur_string += "@@";
  }

  <<EOF>>  unexpected_eof(m_code_start, "}"); BEGIN(INITIAL);
}


  /*--------------------------------------------------------------.
  | Scanning some prologue: from "%{" (already scanned) to "%}".  |
  `--------------------------------------------------------------*/

<SC_PROLOGUE>
{
  "%}" {
    STRING_FINISH;
    BEGIN(INITIAL);
    return make_string_token(m_code_start, last_string, PROLOGUE);
  }

  <<EOF>>  unexpected_eof(m_code_start, "%}"); BEGIN(INITIAL);
}


  /*---------------------------------------------------------------.
  | Scanning the epilogue (everything after the second "%%", which |
  | has already been eaten).                                       |
  `---------------------------------------------------------------*/

<SC_EPILOGUE>
{
  <<EOF>> {
	trace << "Reached EOF in epilogue." << endl;
	STRING_FINISH;
	BEGIN(INITIAL);
	return make_string_token(m_code_start, last_string, EPILOGUE);
  }
}


  /*-----------------------------------------.
  | Escape M4 quoting characters in C code.  |
  `-----------------------------------------*/

<SC_COMMENT,SC_LINE_COMMENT,SC_STRING,SC_CHARACTER,SC_BRACED_CODE,SC_PROLOGUE,SC_EPILOGUE>
{
  \$	cur_string += "$";
  \@	cur_string += "@";
  \[	cur_string += "[";
  \]	cur_string += "]";
}


  /*-----------------------------------------------------.
  | By default, grow the string obstack with the input.  |
  `-----------------------------------------------------*/

<SC_COMMENT,SC_LINE_COMMENT,SC_BRACED_CODE,SC_PROLOGUE,SC_EPILOGUE,SC_STRING,SC_CHARACTER,SC_ESCAPED_STRING,SC_ESCAPED_CHARACTER>.	|
<SC_COMMENT,SC_LINE_COMMENT,SC_BRACED_CODE,SC_PROLOGUE,SC_EPILOGUE>\n	{
	trace << "Now in state " << lexmap[YY_START] << endl;
	STRING_GROW;
}
%%

namespace v3c {
namespace bisonx {
namespace impl {

#define ADD_STATE(x) lexmap[x] = #x
struct lexmap_init {
	lexmap_init()
	{
		ADD_STATE(SC_COMMENT);
		ADD_STATE(SC_LINE_COMMENT);
		ADD_STATE(SC_YACC_COMMENT);
		ADD_STATE(SC_STRING);
		ADD_STATE(SC_CHARACTER);
		ADD_STATE(SC_AFTER_IDENTIFIER);
		ADD_STATE(SC_ESCAPED_STRING);
		ADD_STATE(SC_ESCAPED_CHARACTER);
		ADD_STATE(SC_PRE_CODE);
		ADD_STATE(SC_BRACED_CODE);
		ADD_STATE(SC_PROLOGUE);
		ADD_STATE(SC_EPILOGUE);
	}
#undef ADD_STATE
} lexmap_init_singleton;

/* If BUF is null, add BUFSIZE (which in this case must be less than
   INT_MAX) to COLUMN; otherwise, add mbsnwidth (BUF, BUFSIZE, 0) to
   COLUMN.  If an overflow occurs, or might occur but is undetectable,
   return INT_MAX.  Assume COLUMN is nonnegative.  */

inline int
Scanner::add_column_width(int column, char const *buf, size_t bufsize)
{
  size_t width;
  unsigned int remaining_columns = INT_MAX - column;

  if (buf)
    {
      if (INT_MAX / 2 <= bufsize)
	return INT_MAX;
      width = bisonx::impl::mbsnwidth(buf, bufsize, 0);
    }
  else
    width = bufsize;

  return width <= remaining_columns ? column + width : INT_MAX;
}

/*------------------------------------------------------------------.
| TEXT is pointing to a wannabe semantic value (i.e., a `$').       |
|                                                                   |
| Possible inputs: $[<TYPENAME>]($|integer)                         |
|                                                                   |
`------------------------------------------------------------------*/
inline bool
Scanner::handle_action_dollar()
{
	cur_string += string(yytext, yyleng);
	val->asCode().add_marker(cur_string.length() - yyleng, yyleng);
	return true;
}

/*----------------------------------------------------------------.
| Map `$?' onto the proper M4 symbol, depending on its TOKEN_TYPE |
| (are we in an action?).                                         |
`----------------------------------------------------------------*/
void
Scanner::handle_dollar()
{
	TraceWrap(*this << " : " << __PRETTY_FUNCTION__);

	switch (m_token_type) {
	case BRACED_CODE:
		if (handle_action_dollar())
			return;
		break;
	case PERCENT_DESTRUCTOR:
	case PERCENT_INITIAL_ACTION:
	case PERCENT_PRINTER:
		if (yytext[1] == '$') {
			cur_string += "$$";
			val->asCode().add_marker(cur_string.length() - yyleng, yyleng);
			return;
		}
		break;
	default:
		break;
	}
	complain_at(v3c::Span(*this, yyleng), _("invalid value: %s"), quote(yytext));
}


/*------------------------------------------------------.
| TEXT is a location token (i.e., a `@...').  Output to |
| OBSTACK_FOR_STRING a reference to this location.      |
`------------------------------------------------------*/
bool
Scanner::handle_action_at()
{
	TraceWrap(*this << " : " << __PRETTY_FUNCTION__);

	val->code_value().push_back(Marker(cur_string.length(), yyleng));
	cur_string += string(yytext, yyleng);
	return true;
}


/*----------------------------------------------------------------.
| Map `@?' onto the proper M4 symbol, depending on its TOKEN_TYPE |
| (are we in an action?).                                         |
`----------------------------------------------------------------*/
void
Scanner::handle_at()
{
	TraceWrap(*this << " : " << __PRETTY_FUNCTION__);

	switch (m_token_type) {
	case BRACED_CODE:
		handle_action_at();
		return;
	case PERCENT_INITIAL_ACTION:
	case PERCENT_DESTRUCTOR:
	case PERCENT_PRINTER:
		if (yytext[1] == '$') {
			cur_string += "]b4_at_dollar[";
			return;
		}
		break;
	default:
		break;
	}
	complain_at(v3c::Span(*this, yyleng), _("invalid value: %s"), quote(yytext));
}


/*------------------------------------------------------.
| Scan NUMBER for a base-BASE integer at location LOC.  |
`------------------------------------------------------*/

NodePtr
Scanner::scan_integer(int base, int token)
{
	TraceWrap(*this << " : " << __PRETTY_FUNCTION__);

	verify(INT_MAX < ULONG_MAX);
	errno = 0;
	unsigned long int num = strtoul(yytext, 0, base);

	if (INT_MAX < num) {
		complain_at(v3c::Span(*this, yyleng), _("integer out of range: %s"), quote (yytext));
		num = INT_MAX;
	}
	else if(errno) {
		string s = string_error(errno);
		complain_at(v3c::Span(*this, yyleng), _("integer out of range: %s"), s.c_str());
		num = INT_MAX;
	}
	NodePtr ret = make_int_token(v3c::Span(*this, yyleng), num, token);
	return ret;
}


/*------------------------------------------------------------------.
| Convert universal character name UCN to a single-byte character,  |
| and return that character.  Return -1 if UCN does not correspond  |
| to a single-byte character.					    |
`------------------------------------------------------------------*/
int
Scanner::convert_ucn_to_byte (const char * ucn)
{
  verify (UCHAR_MAX <= INT_MAX);
  unsigned long int code = strtoul (ucn + 2, NULL, 16);

  /* FIXME: Currently we assume Unicode-compatible unibyte characters
     on ASCII hosts (i.e., Latin-1 on hosts with 8-bit bytes).  On
     non-ASCII hosts we support only the portable C character set.
     These limitations should be removed once we add support for
     multibyte characters.  */

  if (UCHAR_MAX < code)
    return -1;

#if ! ('$' == 0x24 && '@' == 0x40 && '`' == 0x60 && '~' == 0x7e)
  {
    /* A non-ASCII host.  Use CODE to index into a table of the C
       basic execution character set, which is guaranteed to exist on
       all Standard C platforms.  This table also includes '$', '@',
       and '`', which are not in the basic execution character set but
       which are unibyte characters on all the platforms that we know
       about.  */
    static signed char const table[] =
      {
	'\0',   -1,   -1,   -1,   -1,   -1,   -1, '\a',
	'\b', '\t', '\n', '\v', '\f', '\r',   -1,   -1,
	  -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
	  -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
	 ' ',  '!',  '"',  '#',  '$',  '%',  '&', '\'',
	 '(',  ')',  '*',  '+',  ',',  '-',  '.',  '/',
	 '0',  '1',  '2',  '3',  '4',  '5',  '6',  '7',
	 '8',  '9',  ':',  ';',  '<',  '=',  '>',  '?',
	 '@',  'A',  'B',  'C',  'D',  'E',  'F',  'G',
	 'H',  'I',  'J',  'K',  'L',  'M',  'N',  'O',
	 'P',  'Q',  'R',  'S',  'T',  'U',  'V',  'W',
	 'X',  'Y',  'Z',  '[', '\\',  ']',  '^',  '_',
	 '`',  'a',  'b',  'c',  'd',  'e',  'f',  'g',
	 'h',  'i',  'j',  'k',  'l',  'm',  'n',  'o',
	 'p',  'q',  'r',  's',  't',  'u',  'v',  'w',
	 'x',  'y',  'z',  '{',  '|',  '}',  '~'
      };

    code = code < sizeof table ? table[code] : -1;
  }
#endif

  return code;
}


/*----------------------------------------------------------------.
| Handle `#line INT "FILE"'.  ARGS has already skipped `#line '.  |
`----------------------------------------------------------------*/

void
Scanner::handle_syncline(char *args)
{
	char *after_num;
	unsigned long int lineno = strtoul (args, &after_num, 10);
	char *szfile = strchr (after_num, '"') + 1;
	*strchr (szfile, '"') = '\0';
	if (INT_MAX <= lineno) {
		warn_at(v3c::Span(*this, yyleng), "%s", _("line number overflow"));
		lineno = INT_MAX;
	}
	filename = szfile;
	line = lineno;
	column = 1;
}


/*---------------------------------.
| Report a rule that is too long.  |
`---------------------------------*/

void
Scanner::rule_length_overflow()
{
  fatal_at(v3c::Span(m_rule_start, *this), "%s", _("rule is too long"));
}


/*----------------------------------------------------------------.
| For a token or comment starting at START, report message MSGID, |
| which should say that an end marker was found before		  |
| the expected TOKEN_END.					  |
`----------------------------------------------------------------*/

void
Scanner::unexpected_end(const v3c::Position & start, char const *msgid, const char * token_end)
{
	complain_at(v3c::Span(start, *this), _(msgid), token_end);
}


/*------------------------------------------------------------------------.
| Report an unexpected EOF in a token or comment starting at START.       |
| An end of file was encountered and the expected TOKEN_END was missing.  |
`------------------------------------------------------------------------*/

void
Scanner::unexpected_eof(const v3c::Position & start, const char * token_end)
{
  unexpected_end(start, N_("missing `%s' at end of file"), token_end);
}


/*----------------------------------------.
| Likewise, but for unexpected newlines.  |
`----------------------------------------*/

void
Scanner::unexpected_newline(const v3c::Position & start, const char * token_end)
{
  unexpected_end(start, N_("missing `%s' at end of line"), token_end);
}


Scanner::Scanner(Parser & parser)
: v3c::Flexx()
, ParserTokens()
, m_parser(parser)
, m_rule_length()
, m_braces_level()
, m_context_state()
, m_token_type()
, m_id_loc()
, m_code_start()
, m_token_start()
, m_bWarningIssued(false)
, m_bComplaintIssued(false)
, max_left_semantic_context()
, last_braced_code_loc()
, last_string()
, cur_string()
, val(parser.yylval)
, m_rule_start()
{
}
/// Assumes that the token's length is yyleng
NodePtr Scanner::make_token(int t)
{
	Node * p = new LexToken(*this, yyleng, t);
	return p;
}
/// Assumes that the token's length is 1
NodePtr Scanner::make_token(const v3c::Position & loc, int t)
{
	Node * p = new LexToken(loc, 1, t);
	return p;
}
NodePtr Scanner::make_string_token(const v3c::Position & loc, const string & text, int token)
{
	TraceWrap(__PRETTY_FUNCTION__);
	Node * p;
	if(&loc == this) { // Compensate for current location.
		p = new StringToken(Span(loc - Offset(text), loc), text, token);
	}else{
		p = new StringToken(loc, text, token);
	}
	trace << "Created token " << *p << endl;
	return p;
}
NodePtr Scanner::make_int_token(const v3c::Span & span, int n, int token)
{
	Node * p = new IntToken(span, n, token);
	return p;
}

#ifdef UNUSED
LexTokenPtr &
Parser::symbol_get(const v3c::Span & s, const string & key)
{
	if(m_symtab.find(key) != m_symtab.end())
		return m_symtab[key];
	SymbolToken * p = new SymbolToken(s, key);
	return m_symtab[key] = p;
}

void Scanner::class_set(LexTokenPtr & p, SymbolClass c, bool declaring)
{
	Symbol & sym = p->asSymbol();
  if (sym.m_class != unknown_sym && sym.m_class != c)
    {
      complain_at(v3c::Span(*this, yyleng), _("symbol %s redefined"), sym->tag);
      sym->declared = false;
    }

  if (class == nterm_sym && sym->class != nterm_sym)
    sym->number = nvars++;
  else if (class == token_sym && sym->number == NUMBER_UNDEFINED)
    sym->number = ntokens++;

  sym->class = class;

  if (declaring)
    {
      if (sym->declared)
	warn_at(v3c::Span(*this, yyleng), _("symbol %s redeclared"), sym->tag);
      sym->declared = true;
    }
}
LexTokenPtr &
Parser::symbol_get(const v3c::Position & loc, const string & key)
{
	if(m_symtab.find(key != m_symtab.end))
		return m_symtab[key];
	SymbolToken * p = new SymbolToken(loc, key);
	return m_symtab[key] = p;
}
#endif
void
Scanner::complain_at(const v3c::Span & loc, const char *message, ...)
{
	va_list args;

	va_start(args, message);
	cerr << "warning: " << loc << ": " << string_inline_vprintf(message, args)
	<< endl;
	va_end(args);

	m_bComplaintIssued = true;
}
void
Scanner::warn_at(const v3c::Span & loc, const char *message, ...)
{
	va_list args;

	va_start (args, message);
	cerr << loc << ": " << string_inline_vprintf(message, args) << endl;
	va_end (args);

	m_bWarningIssued = true;
}

void
fatal_at(const v3c::Span & loc, const char *message, ...)
{
	va_list args;
	stringstream ss;

	va_start (args, message);
	ss << "fatal: " << loc << ": " << string_inline_vprintf(message, args);
	va_end (args);

	throw scanner_error(ss.str());
}

} // namespace impl
} // namespace bisonx
} // namespace v3c
