#if !defined(v3c_bisonx_token_hpp_included)
#define v3c_bisonx_token_hpp_included

// Copyright (C) 2009  Philip Ashmore (contact@philipashmore.com)
// License: LGPLv3.  See LICENSE.txt for the full license.

#include <set>
#include <vector>
#include <string>
#include <stdexcept>
#include <v3c/smart_ptr.hpp>
#include <v3c/span.hpp>

namespace v3c {
namespace bisonx {
namespace impl {

using namespace std;

class Node;
typedef smart_ptr< Node, SingleThreadModelTraits > NodePtr;
ostream & operator <<(ostream & s, const NodePtr & t);
typedef vector< NodePtr > NodePtrVec;
ostream & operator <<(ostream & s, const NodePtrVec & t);

class LexToken;    // A lexical token without any semantic value.
class IntToken;    // A lexical token with an integer semantic value.
class StringToken; // A lexical token with a text semantic value.
class CodeToken;   // A lexical token with a text semantic value and variable
                   // markers.

class Token;
class TokenList;
class Rule;
class Symbol;
class SymbolRef;
class Production;

enum Assoc
{ undef_assoc
, right_assoc
, left_assoc
, non_assoc
, invalid_assoc
};
inline const char * to_char_ptr(Assoc a);

struct Marker
{
	Marker(unsigned int offset_, unsigned int length_)
	: offset(offset_), length(length_)
	{}
	unsigned int offset, length;
};
typedef vector< Marker > MarkerVec;

/**
The LexToken class and it's descendents is designed to replace the equivalent C
union
\code
%union
{
  symbol *symbol;
  symbol_list *list;
  int integer;
  char *chars;
  assoc assoc;
  uniqstr uniqstr;
};
\endcode
symbol         -> SymbolToken
symbol_list    -> [handled by the grammar]
int            -> IntToken
uniqstr,char * -> StringToken
assoc          -> IntToken
*/
class Node : public v3c::Span
{
public:
	enum Type {
	// Lexical structures
	  eLexToken
	, eInt, eString, eCode
	// Semantic structures
	, eToken, eTokenList, eRule, eSymbolRef
	};
	inline Node(const v3c::Position & l, unsigned int length, Type t);
	inline Node(const v3c::Span & s, Type t);
	inline Node(Type t);
	inline virtual ~Node(); // The one and only virtual function.

	inline v3c::Span & asSpan();
	inline const v3c::Span & asSpan() const;

	inline Type type() const;
	inline const char * type_name() const;

	inline LexToken & asLexToken();
	inline const LexToken & asLexToken() const;

	inline IntToken & asInt();
	inline const IntToken & asInt() const;

	inline StringToken & asString();
	inline const StringToken & asString() const;

	inline CodeToken & asCode();
	inline const CodeToken & asCode() const;

	inline Symbol & asSymbol();
	inline const Symbol & asSymbol() const;

	inline Token & asToken();
	inline const Token & asToken() const;

	inline TokenList & asTokenList();
	inline const TokenList & asTokenList() const;

	inline Rule & asRule();
	inline const Rule & asRule() const;

	inline SymbolRef & asSymbolRef();
	inline const SymbolRef & asSymbolRef() const;

	inline void id_set(int id_);
	inline int id() const;

	inline int & int_value();
	inline int int_value() const;

	//inline string & string_value();
	inline string string_value() const;

	inline MarkerVec & code_value();
	inline const MarkerVec & code_value() const;

	inline bool is_symbol() const;
	inline bool is_token() const;
	inline bool is_rule() const;
	inline bool is_string() const;
protected:
	Type m_type;
};
typedef vector< NodePtr > NodePtrVec;

ostream & operator <<(ostream & s, const Node & t);
inline ostream & operator <<(ostream & s, const NodePtr & t);

class LexToken : public Node
{
public:
	inline LexToken(const v3c::Position & l, unsigned int length, int id_
		, Type t = eLexToken);
	inline LexToken(const v3c::Span & s, int id_, Type t = eLexToken);

	int m_nId;
};

/// The scanners semantic value
typedef smart_ptr< LexToken, SingleThreadModelTraits > LexTokenPtr;

class IntToken : public LexToken
{
public:
	inline IntToken(const v3c::Position & p, unsigned int length, int v, int id);
	inline IntToken(const v3c::Span & s, int v, int id);

	int value;
};

class StringToken : public LexToken
{
public:
	inline StringToken(const v3c::Position & p, const string & s, int id
		, Node::Type t = eString);
	inline StringToken(const v3c::Position & p, const char * s, unsigned int len
		, int id, Node::Type t = eString);
	inline StringToken(const v3c::Span & s, const string & txt, int id
		, Node::Type t = eString);

	string value;
};

class CodeToken : public StringToken
{
public:
	inline CodeToken(const v3c::Position & p, const string & s, int id);
	inline CodeToken(const v3c::Position & p, const char * s, unsigned int len
		, int id);
	inline CodeToken(const v3c::Span & s, const string & txt, int id);

	inline void add_marker(unsigned int offs, unsigned int len);
	MarkerVec values; // $$ $<type>1 etc.
};

} // namespace impl
} // namespace bisonx
} // namespace v3c

#include "bisonx_token.impl.hpp"

#endif // #if !defined(bisonx_token_hpp_included)
