#if !defined(v3c_bisonx_token_impl_hpp_included)
#define v3c_bisonx_token_impl_hpp_included

// Copyright (C) 2009  Philip Ashmore (contact@philipashmore.com)
// License: LGPLv3.  See LICENSE.txt for the full license.

#if !defined(v3c_bisonx_token_hpp_included)
#error This file may only be included from v3c_bisonx_token.hpp
#endif

namespace v3c {
namespace bisonx {
namespace impl {

inline const char * to_char_ptr(Assoc a)
{
	switch (a) {
	default:
		throw runtime_error(string_inline_printf(
			"Unknown Assoc value %u", (unsigned int)a));
	case undef_assoc:
		return "<undefined>";
	case right_assoc:
		return "%right";
	case left_assoc:
		return "%left";
	case non_assoc:
		return "%nonassoc";
	}
}

////////////////////////////////////////////////////////////////////////////////
// Node
////////////////////////////////////////////////////////////////////////////////
inline Node::Node(const v3c::Position & p, unsigned int length, Type t)
: v3c::Span(p, length)
, m_type(t)
{
}
inline Node::Node(const v3c::Span & s, Type t)
: v3c::Span(s)
, m_type(t)
{
}
inline Node::Node(Type t)
: v3c::Span()
, m_type(t)
{
}
/// The one and only virtual function.
inline Node::~Node()
{
}
inline Node::Type Node::type() const
{
	return m_type;
}
inline const char * Node::type_name() const
{
	switch(m_type) {
	case eLexToken:
		return "LexToken";
	case eInt:
		return "Int";
	case eString:
		return "String";
	case eCode:
		return "Code";
	case eToken:
		return "Token";
	case eTokenList:
		return "TokenList";
	case eRule:
		return "Rule";
	case eSymbolRef:
		return "SymbolRef";
#ifndef __clang__
	// All cases covered.
	default:
		return "<bad type>";
#endif
	}
}

inline v3c::Span & Node::asSpan()
{
	return *this;
}
inline const v3c::Span & Node::asSpan() const
{
	return *this;
}
inline LexToken & Node::asLexToken()
{
	if(m_type > eCode)
		throw runtime_error(string("asLexToken called on ") + type_name());
	return static_cast< LexToken & >(*this);
}
inline const LexToken & Node::asLexToken() const
{
	if(m_type > eCode)
		throw runtime_error(string("asLexToken called on ") + type_name());
	return static_cast< const LexToken & >(*this);
}
inline IntToken & Node::asInt()
{
	if(m_type == eInt)
		return static_cast< IntToken & >(*this);
	throw runtime_error(string("asInt called on ") + type_name());
}
inline const IntToken & Node::asInt() const
{
	if(m_type == eInt)
		return static_cast< const IntToken & >(*this);
	throw runtime_error(string("asInt called on ") + type_name());
}
inline StringToken & Node::asString()
{
	if(m_type == eString || m_type == eCode)
		return static_cast< StringToken & >(*this);
	throw runtime_error(string("asString called on ") + type_name());
}
inline const StringToken & Node::asString() const
{
	if(m_type == eString || m_type == eCode)
		return static_cast< const StringToken & >(*this);
	throw runtime_error(string("asString called on ") + type_name());
}
inline CodeToken & Node::asCode()
{
	if(m_type == eCode)
		return static_cast< CodeToken & >(*this);
	throw runtime_error(string("asCode called on ") + type_name());
}
inline const CodeToken & Node::asCode() const
{
	if(m_type == eCode)
		return static_cast< const CodeToken & >(*this);
	throw runtime_error(string("asCode called on ") + type_name());
}

inline void Node::id_set(int id_)
{
	asLexToken().m_nId = id_;
}
inline int Node::id() const
{
	return asLexToken().m_nId;
}
inline int & Node::int_value()
{
	return asInt().value;
}
inline int Node::int_value() const
{
	return asInt().value;
}
inline MarkerVec & Node::code_value()
{
	return asCode().values;
}
inline const MarkerVec & Node::code_value() const
{
	return asCode().values;
}
inline bool Node::is_symbol() const
{
	return m_type == eToken || m_type == eRule;
}
inline bool Node::is_token() const
{
	return m_type == eToken;
}
inline bool Node::is_rule() const
{
	return m_type == eRule;
}
inline bool Node::is_string() const
{
	return m_type == eString || m_type == eSymbolRef;
}

inline ostream & operator <<(ostream & s, const NodePtr & t)
{
	if(t)
		s << *t;
	else
		s << "<null>";
	return s;
}

////////////////////////////////////////////////////////////////////////////////
// LexToken
////////////////////////////////////////////////////////////////////////////////
inline LexToken::LexToken(const v3c::Position & p, unsigned int length, int id_, Type t)
: Node(p, length, t)
, m_nId(id_)
{
}
inline LexToken::LexToken(const v3c::Span & span, int id_, Type t)
: Node(span, t)
, m_nId(id_)
{
}

////////////////////////////////////////////////////////////////////////////////
// IntToken
////////////////////////////////////////////////////////////////////////////////
inline IntToken::IntToken(const v3c::Position & p, unsigned int length, int v, int id)
: LexToken(p, length, id, eInt)
, value(v)
{}
inline IntToken::IntToken(const v3c::Span & s, int v, int id)
: LexToken(s, id, eInt)
, value(v)
{}

////////////////////////////////////////////////////////////////////////////////
// StringToken
////////////////////////////////////////////////////////////////////////////////
inline StringToken::StringToken(const v3c::Position & p, const string & s
, int id, Node::Type t)
: LexToken(Span(p, p + Offset(s)), id, t)
, value(s)
{}
inline StringToken::StringToken(const v3c::Position & p, const char * s
, unsigned int len, int id, Node::Type t)
: LexToken(Span(p, p + Offset(s, len)), id, t)
, value(s, len)
{}
inline StringToken::StringToken(const v3c::Span & s, const string & txt, int id
, Node::Type t)
: LexToken(s, id, t)
, value(txt)
{}

////////////////////////////////////////////////////////////////////////////////
// CodeToken
////////////////////////////////////////////////////////////////////////////////
inline CodeToken::CodeToken(const v3c::Position & p, const string & s, int id)
: StringToken(p, s, id, eCode)
, values()
{}
inline CodeToken::CodeToken(const v3c::Position & p, const char * s
, unsigned int len, int id)
: StringToken(Span(p, p + Offset(s, len)), string(s, len), id, eCode)
, values()
{}
inline CodeToken::CodeToken(const v3c::Span & s, const string & txt
, int id)
: StringToken(s, txt, id, eCode)
, values()
{}
inline void CodeToken::add_marker(unsigned int offs, unsigned int len)
{
	values.push_back(Marker(offs, len));
}

} // namespace impl
} // namespace bisonx
} // namespace v3c

#endif // #if !defined(v3c_bisonx_token_impl_hpp_included)
