// Copyright (C) 2009  Philip Ashmore (contact@philipashmore.com)
// License: LPGLv3.  See LICENSE.txt for the full license.

/**
@mainpage parse - c++ flex enhancement (flexx) and bison alternative (bisonx)
@author Philip Ashmore
@version $(VERSION)

@section copying Copying

Copyright (C) 2009 Philip Ashmore

This library is distributed under LGPL. See the COPYING file for
details.
*/

/*! \page bisonx bisonx
A grammar contains definitions of terminal tokens and non-terminal tokens.

A parser when created is given a start rule and a grammar to process.

Non-terminal tokens, or rules are defined in terms of alternative sequences of
tokens called productions.

When a parser reads a token from the scanner it pushes it onto the semantic list
of all paths, giving them the opportunity to process the token.

As a parser processes a production it shifts the result of each element.

When the parser has processed a production completely it performs a reduction,
forming the rules semantic value, potentially for use as part of some other
rules production as a complete grammar instance.

Reductions may have an associated action. This can be the default action
($$ = $1) or may be user-defined.

Mid-rule actions in a grammar are implemented by splitting the production into a
new rule referenced by the original rule such that the mid-rule action becomes
the "reduce" action for the (anonymous) rule.

\section prec Prescedence
Prescedence controls how rules are grouped to form other rules, deciding
whether to shift or reduce.
This is static for terminal symbols (%%left, %%right and %%nonassoc) and can be
context-dependent for rules using contextual precedence by referencing the
precedence of a possibly unused terminal token using %%prec.

Dynamic precedence (%dprec) specifies which of the surving parsers semantic
value is chosen.

Merge functions allow a user-defined means of resolving multiple valid parses of
a sub-grammar.

Bison uses the precedence of a rule to decide whether to shift or reduce when
either is valid.

BisonX follows all paths the precedence rules don't eliminate.

\subsection conflicts Conflicts
Shift-reduce conflicts happen when two or more productions are valid, some
wanting to shift, others wanting to reduce. The "if-then-else" construct is a
good example.

Reduce-reduce conflicts happen when two or more productions are valid and they
all want to reduce.

These are all handled by %dprec, %merge or %prec.

As a bisonX extension, productions may have relative precedences defined by
production labels.

A production can have a label section at it's end, bounded by ().

Within this section there may be a label to the left of a ':'.

After the ':' there may be one or more labels indicating the other productions
in this rule that are lower than this one.

This is similiar to how a makefile indicates which actions come first.

Productions only need a label if they are needed to define the precedence of
other productions.

Production labels are local to the enclosing rule.

\code
rule :
  (N :) // Nothing
| a (A : N) // A > N
| b (: A)   // > A
| c (C :)   // == N
| d (: N A C) // > N, > A, > C
;
\endcode

\subsection errors Error handling
The special token "error" can form part of a production of a rule in which case
it matches any unexpected token and consumes tokens until the succeeding token
is read.
Like any other token it can have an action associated with it.

This action may call one of
 -# yyerrok - allow error reporting to resume immediately
 -# yyclearin - discard the look-ahead token

By default errors are not reported until some number of tokens have been
successfully processed after a previous error.

\subsection locations Locations
Semantic values have a location (a start and end position).
Tokens have a location.
Therefore the default action ($$ = $1) assigns the location as well as the value.

\section grammar_storage Grammar storage
Grammars are represented using a mnemonic syntax like assembly language code,
running on a grammar virtual machine.
Mnemonics are represented as one or more bytes of data, as follows:
 -# [0] instruction
    what task this instruction is to perform
 -# [1+] instruction-dependent data
The instructions are detailed as follows

\subsection error error : failure with error recovery
length 6
<table>
<tr><th>byte</th><th>value</th><th>name</th><th>comments</th>
<tr><th>0</th><td>00</td><td>id</td><td>error</td></tr>
<tr><th>1</th><td>X</td><td>pad</td></tr>
<tr><th>2-3</th><td>NN</td><td>text</td><td>index of error text (may end up indexing multiple tables for i18n)</td></tr>
<tr><th>4-5</th><td>NN</td><td>symbol</td><td>"skip-to" symbol for recovery.<br>
The error state doesn't finish until tokens have been discarded such that
the skip-to token has been matched.<br>
This is conceptually achieved by a call + discard in a try block in a
while loop, however input mis-alignment would in general require that
the skip-to symbol parse attempt should be restarted with only one scanner
token discard between attempts.<br>
Are both approaches correct?</td></tr>
</table>

\subsection production production : match a sequence of symbol(s) and perform a
reduction.
length (n+2)*2
<table>
<tr><th>byte</th><th>value</th><th>name</th><th>comments</th>
<tr><th>0</th><td>01</td><td>id</td><td>production</td></tr>
<tr><th>1</th><td>X</td><td>pad</td></tr>
<tr><th>2-3</th><td>NN</td><td>action</td><td>action index</td></tr>
<tr><th>4-5</th><td>NN</td><td>symbol 0</td></tr>
<tr><th>6-7</th><td>NN</td><td>symbol 1</td></tr>
<tr><th>...</th><td>NN</td><td>symbol ...</td></tr>
<tr><th>(n+1)*2 - (n+1)*2+1</th><td>NN</td><td>symbol (n-1)</td></tr>
</table>
\subsection rule rule : explore n productions
length (n+2)*2
<table>
<tr><th>byte</th><th>value</th><th>name</th><th>comments</th>
<tr><th>0</th><td>02</td><td>id</td><td>fork</td></tr>
<tr><th>1</th><td>N</td><td>n</td><td>number of paths to create</td></tr>
<tr><th>2-3</th><td>NN</td><td>merge handler</td><td>merge handler index</td></tr>
<tr><th>4-5</th><td>NN</td><td>symbol 0</td></tr>
<tr><th>6-7</th><td>NN</td><td>symbol 1</td></tr>
<tr><th>...</th><td>NN</td><td>symbol ...</td></tr>
<tr><th>(n+1)*2 - (n+1)*2+1</th><td>NN</td><td>symbol (n-1)</td></tr>
</table>
The merge function has the following signature:
\code
typedef Array< Path * > PathArray;
int mergefunc(const PathArray & a);
\endcode
The parent path calls into the first path after creating n-1 additional paths.

Of course if the parent path succeeded then it will also be in the array.

Q: What happens if different paths merge after different scanner tokens?

A: The merge function returns an integer specifying which merge "won", or -1 for
"merge not possible". Of course it is free to just fail the parse if it wants
to.

A "marge not possible" result fails this parse thread.

Given a valid merge index result the parser then continues past that merge,
using the semantic list of the winning path.
\subsection repeat repeat : recurse by iteration
length 10
<table>
<tr><th>byte</th><th>value</th><th>name</th><th>comments</th>
<tr><th>0</th><td>04</td><td>id</td><td>repeat</td></tr>
<tr><th>1</th><td>X</td><td>pad</td></tr>
<tr><th>2-3</th><td>NN</td><td>symbol</td></tr>
<tr><th>4-5</th><td>NN</td><td>setup</td><td>action index</td></tr>
<tr><th>6-7</th><td>NN</td><td>match</td><td>action index</td></tr>
<tr><th>8-9</th><td>NN</td><td>no match</td><td>action index</td></tr>
</table>
Setup typically pushes a semantic value to store the accumulated result.
The match action updates the result with the matched symbol.
An optional "no match" action can do something if no symbol matches are
collected.

\section grammar_generation Generating the grammar
Parsing the grammar file yields a grammar data structure representing the
grammar.

The "Entry point" is the start rule that contains one or more alternate paths
or "productions".

Seperate "paths" are created to explore each production by a "fork" instruction.

Paths are state machines with their own semantic list defining their state.

\section niggles Bison/yacc grammar niggles
\subsection recursion Recursion
The standard way to represent left-recursion is
\code
symbols: symbols symbol { A }
       | / * nothing * /
;
\endcode
but this obfuscates the means to initialise the semantic value for "symbols",
forcing A to check if the list has been initialised already before appending to
it.

\section optimizations Optimizations
\subsection deferral Branch deferral
One or more of productions may contain a common prefix that could be used to
defer a branch, but it would undoubtedly lead to duplication and thus an
increase in the size of the generated grammar.

\subsection tagging Terminal tagging
Before calling a sub-rule in a production, a list of terminal tokens associated
with this path could be accessed to see if any of the sub-rules productions
would ultimately lead to the token returned by the scanner.

A match would yield a list of "goto's" that synthesized the calls needed to end
up at the location(s) after where that token was just accepted by the
production(s).

But once again this would mean an increase in the size of the generated grammar,
at the expense of some hopping around by the paths.
*/
