// Copyright (C) 2009  Philip Ashmore (contact@philipashmore.com)
// License: LPGLv3.  See LICENSE.txt for the full license.

/*! \page bisonx_runtime bisonx runtime
\section bisonx_runtime_parse Parse
The parsing process involves the parser calling read() on the scanner to obtain
tokens.
These tokens are processed by the parsers main path - initially set to the start
rule of the grammar.

<b>The root of a grammar is a rule.</b>

\msc
Scanner,Parser,Path;
Scanner<-Parser [label="read()", URL="\ref Scanner::read()"];
Parser->Path [label="process()", URL="\ref Path::process()"];
\endmsc

\section bisonx_runtime_grammar Grammar
\code
struct Grammar
{
	inline const runtime::Node * rule(unsigned int num) const;
	inline const char * name(unsigned int symbol) const;
	uint16_t start_rule, first_rule_index;
	const char ** symbol_names;
	const runtime::Node ** rules;
};
\endcode

Apart from convenience functions, a grammar contains
 -# start_rule - the rule index that is the entry point for the grammar
 -# first_rule_index - any index below this is a token
 -# symbol_names - token/rule names only
 -# rules - rule addresses - for rule N look up rules[N - first_rule_index]
\section bisonx_runtime_path Path
A path is a route through the grammar in response to scanner tokens.

<b>Only rules create paths to explore productions.</b>

A path can have one or more states.

<b>Only productions and repeat-rules can "push" states.</b>

In either case a rule ends up pushing more paths, so a path can have at most
two states.

\subsection bisonx_runtime_path_state PathState
\code
struct PathState {
	const Node * ip;
	unsigned int index; // For productions.
};
\endcode

\subsection bisonx_runtime_node Node
\code
struct Node
{
	enum Type
	{ eRule
	, eProduction
	, eRepeatToken
	, eRepeatRule
	, eErrorToken
	, eErrorRule };
	inline const char * type_name() const;
	inline unsigned int length() const;
	inline const ErrorTokenNode & asErrorToken() const;
	inline const ErrorRuleNode & asErrorRule() const;
	inline const ProductionNode & asProduction() const;
	inline const RuleNode & asRule() const;
	inline const RepeatTokenNode & asRepeatToken() const;
	inline const RepeatRuleNode & asRepeatRule() const;
	uint8_t type, n;
};
\endcode
Node is the base class for all grammar nodes.

Nodes reference each other through the grammars rule table - by index.

Nodes contain
 -# a type, to identify the node
 -# a multi-purpose numeric value "n"

\subsection bisonx_runtime_rule Rule
\code
struct RuleNode : Node {
	int16_t id, mergefunc; // -1 == default.
	uint16_t entries[];
};
\endcode
Rules contain
 -# the id of the rule for ease of debugging
 -# the mergefunc index to merge multiple valid productions
 -# the indeces of the [n] production nodes
\subsection bisonx_runtime_production Production
\code
struct ProductionNode : Node {
	int16_t action; // -1 == default.
	uint16_t entries[];
};
\endcode
A production contains a sequence of node indeces required to match this
production.
 -# the action index to reduce the value stack and produce a result
 -# the indeces of the [n] token/rule nodes
Symbol index 0 means "end of file" \<EOF\>.

\<EOF\> is not matched by the grammar explicitly, but simply means "stream
complete".

Any symbol index under 300 is a character token.

All other symbol indeces are tokens or rules.

The grammar 
\subsection bisonx_runtime_repeat Repeat
A repeat is a wrapper around a rule or a token reference.

It's job is to collect a sequence of results from the processing of it's symbol
reference.

To do this it needs to
 -# initialise a storage semantic value to accumulate the result
 -# handle the case where no input tokens match
 -# "call" on it's symbol reference to handle input tokens
 -# process the semantic value returned by it's symbol reference so as to
    accumulate a result
 -# store a token "replay index" so that if (or when) a call fails it can update
    the read pointer for subsequent symbols in the production

These operations allow for any constraint on the number of matches to be
enforced.
\subsection bisonx_runtime_error Error
Later!
\subsection bisonx_runtime_path Path
A path is a route through the grammar that matches the sequence of input tokens
returned by the scanner.

For a production this involves walking along the symbol references until the
input token is either accepted or rejected.

For a rule, the path creates sub-paths to walk the productions of the rule.

All the sub-paths that succeed are passed to a resolver function that decides
which ones to follow. If more than one compeletes it must again decide which
ones to accept.

\dot
graph g {
#	rankdir=LR;
	nodesep=0;
	ranksep=0;
	node [fontname="FreeSans", fontsize=8, width=0.3, height=0.3];
#	node [shape=octagon];
	exp0 [label="exp"];
	exp00 [label="exp"];
	exp01 [label="exp"];
	exp000 [label="exp"];
	exp001 [label="exp"];
	args [label="arguments"];
	node [shape=box];
	id0 [label="identifier"];
	id1 [label="identifier"];
	id2 [label="identifier"];
	star0 [label="*"];
	plus [label="+"];
#	{ rank=same; exp0; exp00; exp000; id0; a; }
	{ rank=same; exp0; exp00; exp000; id0; a; }
	{ rank=same; plus; exp001; id1; b; }
	{ rank=same; exp01; id2; c; }
	args -- exp0 -- exp00 -- exp000 -- id0 -- a;
	exp000 -- star0 -- exp001 -- id1 -- b;
	exp00 -- plus -- exp01 -- id2 -- c;
}
\enddot
\dot
graph g {
#	rankdir=LR;
	nodesep=0;
	ranksep=0;
	node [fontname="FreeSans", fontsize=8, width=0.3, height=0.3];
#	node [shape=octagon];
	exp0 [label="exp"];
	exp00 [label="exp"];
	exp01 [label="exp"];
	exp000 [label="exp"];
	exp001 [label="exp"];
	assign [label="assignments"];
	node [shape=box];
	node [shape=plaintext];
	star [label="*"];
	plus [label="+"];
	assign -- exp0;
	exp0 -- exp00;
	exp0 -- exp01;
	{ rank=same; exp00; plus; exp01; }
	exp00 -- exp000;
	exp00 -- plus -- exp01;
	{ rank=same; exp000; star; exp001; }
	exp000 -- a;
	exp000 -- star -- exp001;
	exp00 -- exp001;
	exp001 -- b;
	exp01 -- c;
}
\enddot

*/
/*
a := 1 b := 2 c := 3 a + b * c
unit: assignments exp  { store_result($2); };

assignments: assignments assignment {}
           | / * Nothing.  * /        {};

assignment: "identifier" ":=" exp { assign($1, $3); };

%left '+' '-';
%left '*' '/';
exp: exp '+' exp   { $$ = add ($1, $3); }
   | exp '-' exp   { $$ = sub ($1, $3); }
   | exp '*' exp   { $$ = mult($1, $3); }
   | exp '/' exp   { $$ = div ($1, $3); }
   | "identifier"  { $$ = fetch($1); }
   | "number"      { $$ = $1; };
  \dot
\dot
digraph start {
	rankdir=LR;
	nodesep=0.15;
	ranksep=0.35;
	compound=true;
	node [fontname="FreeSans", fontsize=8, shape=plaintext, width=0.3, height=0.3];
	edge [fontname="FreeSans", fontsize=8];
	unit [shape=record];
	assignments [shape=record,label="<f0> assignments | <f1> repeat"];
	assignment [shape=record];
	identifier [shape=octagon];
	assign [shape=octagon,label=":="];
	expref [shape=record,label="exp"];
	list2 [shape=record,label=" { <l20> assignment | b } "];
	list [shape=record,label=" { <l0> a|<l1> := |<l2> exp } "];
	list2:l20->list [weight=0];
	one [shape=record,label="1"];
	list:l2->one;
	value->list2;
	unit->assignments;
#	{ rank=same; path; unit; }
	path->assignments:f1 [weight=0];
	assignments:f1->assignment;
	assignment->identifier->assign->expref;
}
\enddot
digraph example {
	orientation=portrait;
#	rankdir=LR;
	node [fontname="FreeSans", fontsize=7, shape=Mrecord];
	edge [fontname="FreeSans", fontsize=7];
	type;
	int -> type;
	char -> type;
	pointer [ label="{ pointer|type|modifiers }" ];
	array [ label="{ <a> array|type|length\<opt\> }" ];
	pointer -> type;
	array -> type;
	function_parameter [ label="{ function-parameter|index|name|type }" ];
	function_signature [ label="{ function\nsignature|return-type|[parameters]}" ];
	function [ label="{ function|name|signature}" ];
	function -> type;
	function_signature -> type;
	{ rank=same; char; int; pointer; array; function; function_signature; };
	main_sig [ label="{ <m> main-sig|<r> return-type|<1> argc|<2> argv}" ];
	main_sig:m:e -> function_signature [ style=dashed ];
	main [ label="{ <1> main-function|\"main\"|signature }" ];
	main_sig:r:w -> int;
	argc_parameter [ label="{<f> argc-parameter|1|\"argc\"|<t> type }" ];
	argv_parameter [ label="{<f> argv-parameter|2|\"argv\"|<t> type }" ];
	argc_parameter:f:e -> function_parameter [ style=dashed ];
	argv_parameter:f:w -> function_parameter [ style=dashed ];
	argc_parameter:t -> int;
	main_sig:1:w -> argc_parameter;
	main -> main_sig;
	main:1:e -> function [ style=dashed ];
	argv_array [ label="{ <a> argv-array|<t> type}" ];
	char_ptr [ label="{ <a> char-pointer|<t> type}" ];
	char_ptr:t -> char;
	char_ptr:a:w -> pointer [ style=dashed ];
	argv_array:t -> char_ptr:a;
	argv_array:a:w -> array  [ style=dashed ];
	argv_parameter:t -> argv_array;
	main_sig:2 -> argv_parameter;
}
  \enddot
*/
