#if !defined(v3c_bisonx_parser_hpp_included)
#define v3c_bisonx_parser_hpp_included

// Copyright (C) 2009  Philip Ashmore (contact@philipashmore.com)
// License: LGPLv3.  See LICENSE.txt for the full license.

#include <vector>
#include <set>
#include <v3c/trace_scope.hpp>
#include <v3c/string_utils.hpp>
#include <v3c/array.hpp>
#include <v3c/smart_ptr.hpp>
#include <v3c/pretty_print_streambuf.hpp>
#include <iomanip>

namespace v3c {
namespace bisonx {

using namespace std;

struct Grammar;

struct Result : public exception
{
	enum Status { eAccept, eReject, eAbort, eError };

	inline Result(Status s, const string & messageIn = "") throw();
	inline ~Result() throw();
	virtual inline const char * what() const throw();

	Status status;
	string message;
};

namespace runtime {

using namespace std;
using namespace v3c;

struct ProductionNode;
struct RuleNode;
struct RepeatTokenNode;
struct RepeatRuleNode;
struct ErrorTokenNode;
struct ErrorRuleNode;

/*! The Node hierarchy is intended as a convenience for processing the grammar
automaton. All "instructions" are word-aligned, requiring the occasional pad
byte to ease possible alignment problems between architectures when reading
words.
The constructors are actually a misnomer(!) because all instances are merely
helpers for interpreting one big uint8_t array that represents the grammar
automaton.
*/
struct Node
{
	enum Type
	{ eRule
	, eProduction
	, eRepeatToken
	, eRepeatRule
	, eErrorToken
	, eErrorRule };
//private:
	inline Node(Type t, uint8_t n_ = 0)
	: type(t)
	, n(n_)
	{}
//public:
	inline const char * type_name() const;
	inline unsigned int length() const;
	inline const RuleNode * asRule() const;
	inline const ProductionNode * asProduction() const;
	inline const RepeatTokenNode * asRepeatToken() const;
	inline const RepeatRuleNode * asRepeatRule() const;
	inline const ErrorTokenNode * asErrorToken() const;
	inline const ErrorRuleNode * asErrorRule() const;
	uint8_t type, n;
};
struct ErrorTokenNode : Node {
private:
	inline ErrorTokenNode(uint16_t msg_, uint16_t match_)
	: Node(Node::eErrorToken)
	, msg(msg_)
	, match(match_)
	{}
public:
	uint16_t msg, match;
};
struct ErrorRuleNode : Node {
private:
	inline ErrorRuleNode(uint16_t msg_, uint16_t match_)
	: Node(Node::eErrorRule)
	, msg(msg_)
	, match(match_)
	{}
public:
	uint16_t msg, match;
};
/*! Ok, so you can't really construct a ProductionNode as it stands - the
entries are "in-line" and it's length isn't a constant.
*/
struct ProductionNode : Node {
private:
	inline ProductionNode(uint16_t id_, uint8_t n_, int16_t action_, uint16_t * /*entries_*/)
	: Node(eProduction, n_)
	, id(id_)
	, action(action_)
	{}
public:
	uint16_t id;
	int16_t action; // -1 == default.
	uint16_t entries[];
};
/*! Ok, so you can't really construct a RuleNode as it stands - the entries
are "in-line" and it's length isn't a constant.
*/
struct RuleNode : Node {
private:
	inline RuleNode(uint16_t id_, uint8_t n_, int16_t mergefunc_, uint16_t * /*entries_*/)
	: Node(eRule, n_)
	, id(id_)
	, mergefunc(mergefunc_)
	{}
public:
	uint16_t id;
	int16_t mergefunc; // -1 == default.
	uint16_t entries[];
};
struct RepeatTokenNode : Node {
private:
	inline RepeatTokenNode(uint16_t symbol_
		, int16_t setup_action_
		, int16_t match_action_
		, int16_t no_match_action_)
	: Node(eRepeatToken)
	, symbol(symbol_)
	, setup_action(setup_action_)
	, match_action(match_action_)
	, no_match_action(no_match_action_)
	{}
public:
	uint16_t symbol;
	// -1 == default.
	uint16_t setup_action, match_action, no_match_action;
};
struct RepeatRuleNode : Node {
private:
	inline RepeatRuleNode(uint16_t symbol_
		, int16_t setup_action_
		, int16_t match_action_
		, int16_t no_match_action_)
	: Node(eRepeatRule)
	, symbol(symbol_)
	, setup_action(setup_action_)
	, match_action(match_action_)
	, no_match_action(no_match_action_)
	{}
public:
	uint16_t symbol;
	// -1 == default.
	uint16_t setup_action, match_action, no_match_action;
};

/*! The Parser does the parsing by means of paths
*/
struct Path {
	enum ChildState {
		/// Ready for another token, not fully matched yet.
		eReady,
		/// The last token completed the pattern.
		eMatched,
		/// No good.
		eFailed
	};
	enum { eFirstTokenIndex = 300 };
	inline static const char * str(ChildState s);
	virtual inline ~Path(){}
};

template< typename TraitsIn >
class PathT : public Path
{
public:
	typedef TraitsIn Traits;
	typedef PathT< Traits > Path;
	typedef typename Traits::PathPtr PathPtr;
	typedef typename Traits::Parser Parser;
	typedef typename Traits::SemanticValue SemanticValue;
protected:
	inline PathT
	( Parser * parser_
	, Path * parent_
	, unsigned int token_index_ = 0);
public:
	inline virtual ~PathT();
	inline void init(Parser * parser_, Path * parent_, unsigned int token_index_);

	class RulePath : public Path
	{
	public:
		typedef typename Traits::SubPaths SubPaths;
		typedef PathT< Traits > Path;
	
		inline RulePath();
		inline RulePath
		( Parser * parser_
		, Path * parent_
		, const RuleNode * start_node
		, unsigned int token_index_ = 0);
		inline void init
		( Parser * parser_
		, Path * parent_
		, const RuleNode * start_node
		, unsigned int token_index_ = 0);
		virtual inline runtime::Path::ChildState process();
		virtual inline void stream(TreeStreambuf & sb, bool last = true) const;
		virtual inline string self_id() const;
		virtual inline SemanticValue & value() { return m_value; }
		virtual const Node * node() const { return m_pNode; }
	//protected:
		inline void branch(const Node * start_node);
		inline void branch(unsigned int rule);
		inline bool is_recursive(unsigned int symbol) const;
		const RuleNode * m_pNode;
		enum State { eInit, eRunning } state;
		SubPaths paths, want_merge;
		SemanticValue m_value;
	private:
		RulePath(const RulePath &);
		RulePath & operator=(const RulePath &);
	};
	class ProductionPath : public Path
	{
	public:
		inline ProductionPath
		( Parser * parser_
		, Path * parent_
		, const ProductionNode * start_node
		, unsigned int token_index_ = 0);
		virtual inline ChildState process();
		virtual inline ChildState process(SemanticValue & v);
		virtual inline void stream(TreeStreambuf & sb, bool last = true) const;
		virtual inline string self_id() const;
		virtual inline SemanticValue & value() { return m_value_stack.front(); }
		virtual const Node * node() const { return m_pNode; }
	//protected:
		const ProductionNode * m_pNode;
		unsigned int m_uIndex;
		PathPtr m_pRule;
		vector < SemanticValue > m_value_stack;
	private:
		ProductionPath(const ProductionPath &);
		ProductionPath & operator=(const ProductionPath &);
	};
	class RepeatTokenPath : public Path
	{
	public:
		inline RepeatTokenPath
		( Parser * parser_
		, Path * parent_
		, const RepeatTokenNode * start_node
		, unsigned int token_index_ = 0);
		virtual inline ChildState process();
		virtual inline string self_id() const;
		virtual inline SemanticValue & value() { return m_value; }
		virtual const Node * node() const { return m_pNode; }
	protected:
		const RepeatTokenNode * m_pNode;
		bool mbFirstTime;
		SemanticValue m_value;
	private:
		RepeatTokenPath(const RepeatTokenPath &);
		RepeatTokenPath & operator=(const RepeatTokenPath &);
	};
	class RepeatRulePath : public Path
	{
	public:
		typedef smart_ptr< RulePath, SingleThreadModelTraits > RulePathPtr;
		inline RepeatRulePath
		( Parser * parser_
		, Path * parent_
		, const RepeatRuleNode * start_node
		, unsigned int token_index_ = 0);
		virtual inline ChildState process();
		virtual inline void stream(TreeStreambuf & sb, bool last = true) const;
		virtual inline string self_id() const;
		virtual inline SemanticValue & value() { return m_value; }
		virtual const Node * node() const { return m_pNode; }
	protected:
		const RepeatRuleNode * m_pNode;
		bool mbFirstTime;
	public:
		RulePathPtr m_pRule;
		SemanticValue m_value;
	private:
		RepeatRulePath(const RepeatRulePath &);
		RepeatRulePath & operator=(const RepeatRulePath &);
	};
	class ErrorTokenPath : public Path
	{
	public:
		inline ErrorTokenPath
		( Parser * parser_
		, Path * parent_
		, const ErrorTokenNode * start_node
		, unsigned int token_index_ = 0);
		virtual inline ChildState process();
		virtual inline string self_id() const;
		virtual const Node * node() const { return m_pNode; }
	protected:
		const ErrorTokenNode * m_pNode;
	private:
		ErrorTokenPath(const ErrorTokenPath &);
		ErrorTokenPath & operator=(const ErrorTokenPath &);
	};
	class ErrorRulePath : public Path
	{
	public:
		inline ErrorRulePath
		( Parser * parser_
		, Path * parent_
		, const ErrorRuleNode * start_node
		, unsigned int token_index_ = 0);
		virtual inline ChildState process();
		virtual inline void stream(TreeStreambuf & sb, bool last = true) const;
		virtual inline string self_id() const;
		virtual const Node * node() const { return m_pNode; }
	protected:
		const ErrorRuleNode * m_pNode;
		PathPtr m_pRule;
	private:
		ErrorRulePath(const ErrorRulePath &);
		ErrorRulePath & operator=(const ErrorRulePath &);
	};
	inline PathPtr make_path( const Node * start_node_
							, unsigned int token_index_);
	inline PathPtr make_path( unsigned int rule);
	inline unsigned int depth() const;

	virtual string self_id() const = 0;
	virtual ChildState process() = 0;

	virtual const Node * node() const = 0;
	virtual inline SemanticValue & value();
	inline uint8_t type() const { return node()->type; }
	inline const char * type_name() const { return node()->type_name(); }

	Parser * parser;
	Path * parent;
	unsigned int token_count_since_error;
	bool clearin; // Discard the next token.
	unsigned int path_id, token_index;
	inline bool isToken(unsigned int symbol) const;
	inline string id() const;
	virtual inline void stream(TreeStreambuf & sb, bool last = true) const;
	inline RulePath & asRule();
	inline ProductionPath & asProduction();
	inline RepeatTokenPath & asRepeatToken();
	inline RepeatRulePath & asRepeatRule();
	inline ErrorTokenPath & asErrorToken();
	inline ErrorRulePath & asErrorRule();
private:
	PathT(const PathT< Traits > & o);
	PathT< Traits > & operator =(const PathT< Traits > & o);
};

template < typename Traits >
class ParserT
{
public:
	typedef typename Traits::PathPtr PathPtr;
	typedef typename Traits::SubPaths SubPaths;
	typedef typename Traits::Scanner Scanner;
	typedef typename Traits::SemanticValue SemanticValue;

	inline ParserT(const Grammar & grammar_);
	inline ~ParserT();
	inline void reset();
	inline unsigned int error_supress_count() const;
	inline int parse(int start_rule = -1);
	inline bool isToken(unsigned int symbol) const;
	inline void stream(ostream & s_) const;
	inline void stream(TreeStreambuf & sb) const;
	inline void dump() const;
	// void onAction(ParserMachine * m, unsigned int index); // 0 == default action.
	// void onMerge();
	const Grammar * grammar;
	Scanner scanner;
	vector < SemanticValue > tokens;
	typename PathT< Traits >::RulePath root;
	inline unsigned int next_path_id();
protected:
	unsigned int m_next_path_id;
private:
	ParserT(const ParserT< Traits > & o);
	ParserT< Traits > & operator =(const ParserT< Traits > & o);
};

} // namespace runtime

/*! A grammar contains the means to parse a language instance, an automaton.

This structure stores:
1. first rule index
2. the symbol names table (tokens + rules)
3. the rules table
*/
struct Grammar
{
	Grammar
	( uint16_t start_rule_
	, uint16_t first_rule_index_
	, const char ** symbol_names_
	, const runtime::Node ** rules_
	)
	: start_rule(start_rule_)
	, first_rule_index(first_rule_index_)
	//, pad()
	, symbol_names(symbol_names_)
	, paths(rules_)
	{}
	inline const runtime::Node * path(unsigned int num) const;
	inline const char * name(unsigned int symbol) const;
	uint16_t start_rule, first_rule_index;
	//uint32_t pad;
	const char ** symbol_names;
	const runtime::Node ** paths;
};

} // namespace bisonx
} // namespace v3c

#include "parser.timpl.hpp"

#endif // #if !defined(v3c_bisonx_parser_hpp_included)
