// Copyright (C) 2009  Philip Ashmore (contact@philipashmore.com)
// License: LGPLv3.  See LICENSE.txt for the full license.

#if !defined(v3c_bisonx_parser_hpp_included)
#error This file may only be included from parser.hpp
#endif

namespace v3c {
namespace bisonx {
////////////////////////////////////////////////////////////////////////////////
// Result
////////////////////////////////////////////////////////////////////////////////
inline Result::Result(Status s, const string & messageIn) throw()
: exception()
, status(s)
, message(messageIn)
{
	if(!messageIn.length())
		switch(s) {
		case eAccept:
			message = "Accepted.";
			break;
		case eReject:
			message = "Rejected.";
			break;
		case eAbort:
			message = "Aborted.";
			break;
		case eError:
			message = "Error.";
			break;
		}
	else
		message = messageIn;
}
inline Result::~Result() throw() {}
inline const char * Result::what() const throw()
{
	return message.c_str();
}

namespace runtime {
////////////////////////////////////////////////////////////////////////////////
// Node
////////////////////////////////////////////////////////////////////////////////
inline const char * Node::type_name() const
{
	switch(type) {
	case eRule:
		return "Rule";
	case eProduction:
		return "Production";
	case eRepeatToken:
		return "RepeatToken";
	case eRepeatRule:
		return "RepeatRule";
	case eErrorToken:
		return "ErrorToken";
	case eErrorRule:
		return "ErrorRule";
	default:
		return "<?>";
	}
}
/*! Not sure this is needed.
*/
inline unsigned int Node::length() const
{
	switch(type) {
	case eRule:
	case eProduction:
		return 6 + n*2;
	case eRepeatToken:
	case eRepeatRule:
		return 10;
	case eErrorToken:
	case eErrorRule:
		return 6;
	default:
		throw runtime_error("Length of unknown m_pNode type is unknown.");
	}
}
inline const RuleNode * Node::asRule() const
{
	if(type == eRule)
		return static_cast< const RuleNode * >(this);
	throw runtime_error(string("asRule called on ") + type_name());
}
inline const ProductionNode * Node::asProduction() const
{
	if(type == eProduction)
		return static_cast< const ProductionNode * >(this);
	throw runtime_error(string("asProduction called on ") + type_name());
}
inline const RepeatTokenNode * Node::asRepeatToken() const
{
	if(type == eRepeatToken)
		return static_cast< const RepeatTokenNode * >(this);
	throw runtime_error(string("asRepeatToken called on ") + type_name());
}
inline const RepeatRuleNode * Node::asRepeatRule() const
{
	if(type == eRepeatRule)
		return static_cast< const RepeatRuleNode * >(this);
	throw runtime_error(string("asRepeatRule called on ") + type_name());
}
inline const ErrorTokenNode * Node::asErrorToken() const
{
	if(type == eErrorToken)
		return static_cast< const ErrorTokenNode * >(this);
	throw runtime_error(string("asErrorToken called on ") + type_name());
}
inline const ErrorRuleNode * Node::asErrorRule() const
{
	if(type == eErrorRule)
		return static_cast< const ErrorRuleNode * >(this);
	throw runtime_error(string("asErrorRule called on ") + type_name());
}
////////////////////////////////////////////////////////////////////////////////
// Path
////////////////////////////////////////////////////////////////////////////////
inline const char * Path::str(ChildState s)
{
	switch(s) {
	case eReady:
		return "ready";
	case eMatched:
		return "matched";
	case eFailed:
		return "failed";
#ifndef __clang__
	// All cases covered.
	default:
		return "<?>";
#endif
	}
}
////////////////////////////////////////////////////////////////////////////////
// PathT
////////////////////////////////////////////////////////////////////////////////
template< typename Traits >
inline PathT< Traits >::PathT(typename Traits::Parser * parser_, PathT * parent_, unsigned int token_index_)
: runtime::Path()
, parser(parser_)
, parent(parent_)
, token_count_since_error(parser->error_supress_count())
, clearin(false) // ?
, path_id(parser ? parser->next_path_id() : 0)
, token_index(token_index_)
{
	TraceWrap("PathT< Traits >::PathT()");
}
template< typename Traits >
inline PathT< Traits >::~PathT()
{
	TraceWrap("PathT< Traits >::~PathT()");
}
template< typename Traits >
inline void PathT< Traits >::init
( Parser * parser_
, Path * parent_
, unsigned int token_index_)
{
	parser = parser_;
	parent = parent_;
	path_id = parser->next_path_id();
	token_index = token_index_;
	token_count_since_error = parser->error_supress_count();
}
/*! Get a token

Things to consider:
 -# we may be left recursive in which case our parent has an instance already
 -# a merge could choose a path that didn't get to the last scanner token
*/
template< typename Traits >
inline typename PathT< Traits >::PathPtr PathT< Traits >::make_path
( const Node * start_node
, unsigned int token_index_)
{
	switch(start_node->type) {
	case Node::eRule:
		return new RulePath(parser, this, start_node->asRule(), token_index_);
	case Node::eProduction:
		return new ProductionPath(parser, this, start_node->asProduction()
								, token_index_);
	case Node::eRepeatToken:
		return new RepeatTokenPath(parser, this, start_node->asRepeatToken()
								, token_index_);
	case Node::eRepeatRule:
		return new RepeatRulePath(parser, this, start_node->asRepeatRule()
								, token_index_);
	case Node::eErrorToken:
		return new ErrorTokenPath(parser, this, start_node->asErrorToken()
								, token_index_);
	case Node::eErrorRule:
		return new ErrorRulePath(parser, this, start_node->asErrorRule()
								, token_index_);
	default:
		throw runtime_error("PathT< Traits >::make_path() : Bad type!");
	}
}
template< typename Traits >
inline typename PathT< Traits >::PathPtr PathT< Traits >::make_path
(unsigned int rule)
{
	const Node * p = parser->grammar->path(rule);
	return make_path(p, token_index);
}
template< typename Traits >
inline unsigned int PathT< Traits >::depth() const
{
	PathT *p = parent;
	unsigned int ret = 0;
	while(p) {
		ret++;
		p = p->parent;
	}
	return ret;
}
template< typename Traits >
inline bool PathT< Traits >::isToken(unsigned int symbol) const
{
	return parser->isToken(symbol);
}
template< typename Traits >
inline string PathT< Traits >::id() const
{
	if(parent)
		//return parent->id() + string("->") + parser->grammar->name(ip->;
		return parent->id() + string("->") + self_id();
	else{
		return self_id();
	}
}
template< typename Traits >
inline void PathT< Traits >::stream(TreeStreambuf & sb, bool last) const
{
	sb.add(string_inline_printf("[%u] path %u ", token_index, path_id) + self_id(), last);
}/*
template< typename Traits >
Path::ChildState PathT< Traits >::process()
{
	TraceWrap("PathT< Traits >::process(" << token_index << ')');
	ChildState res;
	SemanticValue & v = parser->tokens[token_index];
	res = process(v);
	trace << "Returning " << str(res) << '.' << endl;
	return res;
}*/
template< typename Traits >
inline typename PathT< Traits >::RulePath & PathT< Traits >::asRule()
{
	if(type() == Node::eRule)
		return static_cast< typename PathT< Traits >::RulePath & >(*this);
	throw runtime_error(string("asRule called on ") + type_name());
}
template< typename Traits >
inline typename PathT< Traits >::ProductionPath & PathT< Traits >::asProduction()
{
	if(type() == Node::eProduction)
		return static_cast< typename PathT< Traits >::ProductionPath & >(*this);
	throw runtime_error(string("asProduction called on ") + type_name());
}
template< typename Traits >
inline typename PathT< Traits >::RepeatTokenPath & PathT< Traits >::asRepeatToken()
{
	if(type() == Node::eRepeatToken)
		return static_cast< typename PathT< Traits >::RepeatTokenPath & >(*this);
	throw runtime_error(string("asRepeatToken called on ") + type_name());
}
template< typename Traits >
inline typename PathT< Traits >::RepeatRulePath & PathT< Traits >::asRepeatRule()
{
	if(type() == Node::eRepeatRule)
		return static_cast< typename PathT< Traits >::RepeatRulePath & >(*this);
	throw runtime_error(string("asRepeatRule called on ") + type_name());
}
template< typename Traits >
inline typename PathT< Traits >::ErrorTokenPath & PathT< Traits >::asErrorToken()
{
	if(type() == Node::eErrorToken)
		return static_cast< typename PathT< Traits >::ErrorTokenPath & >(*this);
	throw runtime_error(string("asErrorToken called on ") + type_name());
}
template< typename Traits >
inline typename PathT< Traits >::ErrorRulePath & PathT< Traits >::asErrorRule()
{
	if(type() == Node::eErrorRule)
		return static_cast< typename PathT< Traits >::ErrorRulePath & >(*this);
	throw runtime_error(string("asErrorRule called on ") + type_name());
}
template< typename Traits >
inline typename Traits::SemanticValue & PathT< Traits >::value()
{
	throw runtime_error(string("value() called on ") + type_name());
}
////////////////////////////////////////////////////////////////////////////////
// PathT< Traits >::RulePath
////////////////////////////////////////////////////////////////////////////////
template< typename Traits >
inline PathT< Traits >::RulePath::RulePath()
: PathT< Traits >(0, 0, 0)
, m_pNode()
, state(eInit)
, paths()
, want_merge()
, m_value()
{
}
template< typename Traits >
inline PathT< Traits >::RulePath::RulePath
( Parser * parser_
, Path * parent_
, const RuleNode * start_node
, unsigned int token_index_)
: PathT< Traits >(parser_, parent_, token_index_)
, m_pNode(start_node)
, state(eInit)
, paths()
, want_merge()
, m_value()
{
}
template< typename Traits >
inline void PathT< Traits >::RulePath::init
( Parser * parser_
, Path * parent_
, const RuleNode * start_node
, unsigned int token_index_)
{
	PathT<Traits>::init(parser_, parent_, token_index_);
	m_pNode = start_node;
	state = eInit;
	paths.clear();
	want_merge.clear();
	m_value = SemanticValue();
}
template< typename Traits >
inline void PathT< Traits >::RulePath::branch(const Node * start_node)
{
	PathPtr path = make_path(start_node, this->token_index);
	if(!path)
		throw runtime_error("Couldn't create path!");
	paths.push_back(path);
	//path->value_stack.push_back(value_stack.back());
	this->parser->dump();
}
template< typename Traits >
inline void PathT< Traits >::RulePath::branch(unsigned int rule_index)
{
	branch(this->parser->grammar->path(rule_index));
}
template< typename Traits >
inline bool PathT< Traits >::RulePath::is_recursive(unsigned int symbol) const
{
	TraceWrap("RulePath::is_recursive(" << this->parser->grammar->name(symbol)
		<< ')');
	const RulePath * p = this;
	const ProductionNode * path
		= this->parser->grammar->path(symbol)->asProduction();
	unsigned int compare = path->entries[0];
	while(p) {
		bool b1 = (p->token_index == this->token_index);
		bool b2 = (compare == p->m_pNode->id);
		trace << "p->token_index(" << p->token_index << ") == token_index("
			<< this->token_index << ") : " << b1 << ", "
			<< this->parser->grammar->name(compare) << " == "
			<< this->parser->grammar->name(p->m_pNode->id) << " : " << b2 << '.'
			<< endl;
		if(b1 && b2)
			return true; // Haven't consumed any more tokens and is the same type.
		Path * q = p->parent;
		bool repeat;
		do {
			if(!q) {
				p = 0;
				break;
			}
			repeat = false;
			switch(q->type()) {
			case Node::eRule:
				trace << "Rule" << endl;
				p = & q->asRule();
				break;
			case Node::eRepeatRule: {
				trace << "RepeatRule" << endl;
				const RepeatRulePath * r = & q->asRepeatRule();
				if(r->m_pRule) {
					if(p == r->m_pRule.pointer()) {
						// We've been here before.
						q = r->parent;
						repeat = true;
					}else
						p = r->m_pRule.pointer();
				}else
					throw runtime_error("RulePath::is_recursive() : RepeatRule has "
						"no Rule!");
				break; }
			case Node::eProduction:
				q = q->parent; // Checking that it's not the first symbol isn't enough.
				repeat = true;
				break;
			case Node::eRepeatToken: // Shouldn't happen.
			case Node::eErrorToken: // Shouldn't happen.
			case Node::eErrorRule: // Need to figure this one out.
				throw runtime_error(string("RulePath::is_recursive() : TODO : "
					"handle ") + q->type_name());
				break;
			}
		} while(repeat);
	}
	return false;
}
/*! A rule m_pNode details a list of productions

We don't update our token_index until we perform a merge, at which point we use
the token_index of the "winning" production.

TODO allow a rule to split so it can follow multiple merges.
*/
template< typename Traits >
inline Path::ChildState PathT< Traits >::RulePath::process()
{
	TraceWrap("RulePath::process(" << this->path_id << " [" << (uint16_t)m_pNode->n
		<< " production(s)])");
	if(depth() >= 6) {
		trace << "### Avoiding infinite recursion ###" << endl;
		return eFailed; // Debugging/infinite recursion/stop here.
	}
	ChildState res;
	switch(state) {
	case eInit:
		// First time through.
		// Create paths for each production.
		for(unsigned int n = 0; n < m_pNode->n; n++) {
			unsigned int symbol = m_pNode->entries[n];
			if(is_recursive(symbol)) {
				trace << "Not recursing into "
					<< this->parser->grammar->name(symbol) << '.' << endl;
			}else{
				branch(symbol);
			}
		}
		state = eRunning;
		// Fall through.
	case eRunning: {
		// Let the child paths process the tokens.
		SubPaths copy = paths;
		typename SubPaths::iterator i = copy.begin();
		typename SubPaths::iterator j = copy.end();
		for(; i != j; ++i) {
			Path * p = (*i).pointer();
			res = p->process();
			switch(res) {
			case Path::eReady:
				break;
			case Path::eMatched:
				want_merge.push_back(*i); // Preserve smart_ptr
				// Fall through.
			case Path::eFailed: {
				unsigned int n;
				bool erased = false;
				typename SubPaths::iterator x = paths.begin();
				for(n = 0; n < paths.size(); ++n, ++x)
					if((*x).pointer() == (*i).pointer()) {
						paths.erase(x);
						erased = true;
						break;
					}
				if(!erased)
					throw runtime_error("Failed to find path to erase.");
				break; }
			}
		}
		// See if all our sub-paths have finished.
		// Only then can we run the merge func.
		if(!paths.size()) {
			trace << "All paths complete." << endl;
			if(!want_merge.size()) {
				trace << "No matched paths - failing." << endl;
				return eFailed;
			}
			this->parser->onMerge(*this, m_pNode->mergefunc);
			this->token_index = want_merge[0]->token_index;
			want_merge.clear();
			return eMatched;
		}
		break; }
#ifndef __clang__
	// All cases covered.
	default:
		throw runtime_error("Invalid onRule index.");
#endif
	}
	return eReady;
}
template< typename Traits >
inline void PathT< Traits >::RulePath::stream(TreeStreambuf & sb, bool last) const
{
	PathT< Traits >::stream(sb, last);
	sb.push();
	this->parser->stream(sb, m_value, 0, 0, false);
	sb.add("Running", false);
	sb.push();
	typename SubPaths::const_iterator p = paths.begin();
	typename SubPaths::const_iterator q = paths.end();
	unsigned int ulast = paths.size() - 1;
	for(unsigned int n = 0; p != q; ++p, ++n)
		(*p)->stream(sb, n == ulast);
	sb.pop();
	sb.add("Matched", last);
	sb.push();
	p = want_merge.begin();
	q = want_merge.end();
	ulast = want_merge.size() - 1;
	for(unsigned int n = 0; p != q; ++p, ++n)
		(*p)->stream(sb, n == ulast);
	sb.pop();
	sb.pop();
}
template< typename Traits >
inline string PathT< Traits >::RulePath::self_id() const
{
	stringstream ss;
	ss << "Rule [" << this->parser->grammar->name(m_pNode->id) << ']';
	return ss.str();
}
////////////////////////////////////////////////////////////////////////////////
// PathT< Traits >::ProductionPath
////////////////////////////////////////////////////////////////////////////////
template< typename Traits >
inline PathT< Traits >::ProductionPath::ProductionPath
( Parser * parser_
, Path * parent_
, const ProductionNode * start_node
, unsigned int token_index_)
: PathT< Traits >(parser_, parent_, token_index_)
, m_pNode(start_node)
, m_uIndex(0)
, m_pRule()
, m_value_stack()
{
}
/*! See if we start with a rule (left recursion).

If yes then look at our parent rule for a match.
*/
template< typename Traits >
Path::ChildState PathT< Traits >::ProductionPath::process()
{
	TraceWrap("PathT< Traits >::ProductionPath::process()");
	ChildState res;
	// Ok so we're checking things backwards but this should be a better
	// filter.

	if(this->parent) {
		RulePath *rp = 0;
		if(this->parent->type() == Node::eRepeatRule) {
			trace << "Parent is a RepeatRulePath." << endl;
			RepeatRulePath *rrp = & this->parent->asRepeatRule();
			if(rrp->m_pRule) {
				rp = rrp->m_pRule.pointer();
			trace << "... with an active RulePath." << endl;
			}
		}else if(this->parent->type() == Node::eRule) {
			trace << "Parent is a RulePath." << endl;
			rp = & this->parent->asRule();
		}
		if(rp && !m_uIndex && rp->value()) {
			trace << "Got a value from the RulePath." << endl;
			unsigned int symbol = m_pNode->entries[m_uIndex];
			if(symbol == (unsigned)rp->value()->id()) { // And therefore a rule.
				trace << "Matched left-recursion." << endl;
				m_value_stack.push_back(rp->value());
				m_uIndex++;
				if(m_uIndex == m_pNode->n)
					throw runtime_error("Rule contains self only!?!");
			}
		}
	}
	trace << "Getting scanner token " << this->token_index << '.' << endl;
	trace << "Token is " << this->parser->tokens[this->token_index] << '.'
		<< endl;
	res = process(this->parser->tokens[this->token_index]);
	trace << "Returning " << str(res) << '.' << endl;
	return res;
}
/*! A production is a list of symbols.
*/
template< typename Traits >
inline Path::ChildState PathT< Traits >::ProductionPath::process
(PathT< Traits >::SemanticValue & v)
{
	TraceWrap("PathT< Traits >::ProductionPath::process(" << this->path_id
		<< " : " << this->parser->str(v) << ')');

	unsigned int symbol = m_pNode->entries[m_uIndex];
	if(isToken(symbol)) {
		if(symbol != (unsigned int)v->id()) {
			trace << self_id() << " : != " << this->parser->str(v) << '.'
				<< endl;
			return eFailed;
		}
		trace << self_id() << " : matched " << this->parser->str(v) << '.'
			<< endl;
		m_uIndex++;
		this->token_index++;
		m_value_stack.push_back(v);
	}else{
		trace << "Calling rule." << endl;
		if(!m_pRule)
			m_pRule = make_path(symbol);
		ChildState res = m_pRule->process();
		trace << "Called rule returned " << str(res) << '.' << endl;
		switch(res) {
		case eReady:
			break;
		case eMatched:
			this->token_index = m_pRule->token_index;
			m_value_stack.push_back(m_pRule->value());
			m_uIndex++;
			m_pRule = PathPtr();
			break;
		case eFailed:
			return res;
		}
	}
	if(m_uIndex == m_pNode->n) {
		// Perform the action for this production.
		this->parser->onAction(*this, m_pNode->action);
		return eMatched;
	}
	return eReady;
}
template< typename Traits >
inline void PathT< Traits >::ProductionPath::stream(TreeStreambuf & sb, bool last) const
{
	PathT< Traits >::stream(sb, last);
	sb.push();
	typename vector< SemanticValue >::const_iterator i = m_value_stack.begin();
	typename vector< SemanticValue >::const_iterator j = m_value_stack.end();
	unsigned int ulast = m_value_stack.size() - !m_pRule;
	for(unsigned int n = 0; i != j; ++i, ++n)
		this->parser->stream(sb, *i, 0, n, n == ulast);
	if(!m_pRule)
		sb.pop();
	else{
		m_pRule->stream(sb, true);
		sb.pop();
	}
}
template< typename Traits >
inline string PathT< Traits >::ProductionPath::self_id() const
{
	stringstream ss;
	ss << "Production [" << this->parser->grammar->name(m_pNode->id) << ']';
	unsigned int n;
	for(n = 0; n < m_pNode->n; n++) {
		if(m_uIndex == n)
			ss << " . ";
		else
			ss << ' ';
		if(!m_pNode->entries[n])
			ss << "<EOF>";
		else if(m_pNode->entries[n] < eFirstTokenIndex)
			ss << '\'' << (char)m_pNode->entries[n] << '\'';
		else
			ss << this->parser->grammar->name(m_pNode->entries[n]);
	}
	if(m_uIndex == n)
		ss << " .";
	return ss.str();
}
////////////////////////////////////////////////////////////////////////////////
// PathT< Traits >::RepeatTokenPath
////////////////////////////////////////////////////////////////////////////////
template< typename Traits >
inline PathT< Traits >::RepeatTokenPath::RepeatTokenPath
( Parser * parser_
, Path * parent_
, const RepeatTokenNode * start_node
, unsigned int token_index_)
: PathT< Traits >(parser_, parent_, token_index_)
, m_pNode(start_node)
, mbFirstTime(true)
, m_value()
{
}
template< typename Traits >
inline string PathT< Traits >::RepeatTokenPath::self_id() const
{
	stringstream ss;
	ss << "RepeatToken [" << this->parser->grammar->name(m_pNode->symbol)
		<< ']';
	return ss.str();
}
template< typename Traits >
inline Path::ChildState PathT< Traits >::RepeatTokenPath::process()
{
	TraceWrap("RepeatTokenPath::process(" << this->path_id << ')');

	SemanticValue & v = this->parser->tokens[this->token_index];
	if(mbFirstTime)
		this->parser->onAction(*this, m_pNode->setup_action);
	// Ready to try our first token.
	if(m_pNode->symbol != v->id()) {
		if(mbFirstTime) {
			trace << id() << " : repeat token never started." << endl;
			return this->parser->onAction(*this, m_pNode->no_match_action);
		}
		return eMatched; // Still considered a match.
	}
	trace << id() << " : token matched." << endl;
	ChildState res = this->parser->onAction(*this, m_pNode->match_action);
	this->token_index++;
	mbFirstTime = false;
	return res;
}
////////////////////////////////////////////////////////////////////////////////
// PathT< Traits >::RepeatRulePath
////////////////////////////////////////////////////////////////////////////////
template< typename Traits >
inline PathT< Traits >::RepeatRulePath::RepeatRulePath
( Parser * parser_
, Path * parent_
, const RepeatRuleNode * start_node
, unsigned int token_index_)
: PathT< Traits >(parser_, parent_, token_index_)
, m_pNode(start_node)
, mbFirstTime(true)
, m_pRule()
, m_value()
{
}
template< typename Traits >
inline Path::ChildState PathT< Traits >::RepeatRulePath::process
()
{
	TraceWrap("RepeatRulePath::process(" << this->path_id << ')');

	if(mbFirstTime) {
		this->parser->onAction(*this, m_pNode->setup_action);
		PathPtr p = make_path(m_pNode->symbol);
		m_pRule = *(RulePathPtr *) & p;
	}
	ChildState res = m_pRule->process();
	trace << "Called rule returned " << str(res) << '.' << endl;
	switch(res) {
	case eReady:
		break;
	case eMatched:
		this->token_index = m_pRule->token_index;
		this->parser->onAction(*this, m_pNode->match_action);
		m_pRule->m_value = SemanticValue();
		res = eReady;
	case eFailed:
		if(mbFirstTime) {
			trace << id() << " : repeat token never started." << endl;
			this->parser->onAction(*this, m_pNode->no_match_action);
		}
		res = eMatched; // Still considered a match.
	}
	mbFirstTime = false;
	return eReady;
}
template< typename Traits >
inline void PathT< Traits >::RepeatRulePath::stream(TreeStreambuf & sb, bool last) const
{
	PathT< Traits >::stream(sb, last);
	sb.push();
	this->parser->stream(sb, m_value, 0, 0, !m_pRule);
	if(!m_pRule)
		sb.pop();
	else{
		m_pRule->stream(sb, true);
		sb.pop();
	}
}
template< typename Traits >
inline string PathT< Traits >::RepeatRulePath::self_id() const
{
	stringstream ss;
	ss << "RepeatRule [" << this->parser->grammar->name(m_pNode->symbol) << ']';
	return ss.str();
}
////////////////////////////////////////////////////////////////////////////////
// PathT< Traits >::ErrorTokenPath
////////////////////////////////////////////////////////////////////////////////
template< typename Traits >
inline PathT< Traits >::ErrorTokenPath::ErrorTokenPath
( Parser * parser_
, Path * parent_
, const ErrorTokenNode * start_node
, unsigned int token_index_)
: PathT< Traits >(parser_, parent_, token_index_)
, m_pNode(start_node)
{
}
template< typename Traits >
inline string PathT< Traits >::ErrorTokenPath::self_id() const
{
	return string("ErrorToken");
}
template< typename Traits >
inline Path::ChildState PathT< Traits >::ErrorTokenPath::process()
{
	return eFailed; // LATER.
}
////////////////////////////////////////////////////////////////////////////////
// PathT< Traits >::ErrorRulePath
////////////////////////////////////////////////////////////////////////////////
template< typename Traits >
inline PathT< Traits >::ErrorRulePath::ErrorRulePath
( Parser * parser_
, Path * parent_
, const ErrorRuleNode * start_node
, unsigned int token_index_)
: PathT< Traits >(parser_, parent_, token_index_)
, m_pNode(start_node)
, m_pRule()
{
}
template< typename Traits >
inline Path::ChildState PathT< Traits >::ErrorRulePath::process()
{
	return eFailed; // LATER.
}
template< typename Traits >
inline void PathT< Traits >::ErrorRulePath::stream(TreeStreambuf & sb, bool last) const
{
	PathT< Traits >::stream(sb, last);
	if(m_pRule) {
		sb.push();
		m_pRule->stream(sb, true);
		sb.pop();
	}
}
template< typename Traits >
inline string PathT< Traits >::ErrorRulePath::self_id() const
{
	return string("ErrorRule");
}
////////////////////////////////////////////////////////////////////////////////
// ParserT
////////////////////////////////////////////////////////////////////////////////
template< typename Traits >
inline ParserT< Traits >::ParserT(const Grammar & grammar_)
: grammar(& grammar_)
, scanner()
, tokens()
, root()
, m_next_path_id(0)
{
	TraceWrap("ParserT< Traits >::ParserT()");
}
template< typename Traits >
inline ParserT< Traits >::~ParserT()
{
	TraceWrap("ParserT< Traits >::~ParserT()");
}
template< typename Traits >
inline unsigned int ParserT< Traits >::error_supress_count() const
{
	return 3; // A usable default.
}
template< typename Traits >
inline void ParserT< Traits >::reset()
{
	//root.reset();
	tokens.clear();
	// We're not in a position to update the scanner to the input/output stream,
	// never mind the input stream name/line number - the caller must do this.
}
template< typename Traits >
inline int ParserT< Traits >::parse(int start_rule)
{
	unsigned int urule = (unsigned int)start_rule;
	if(start_rule == -1)
		urule = grammar->start_rule;
	TraceWrap("ParserT< Traits >::parse(" << urule << ')');
	reset();
	root.init(static_cast< typename Traits::Parser * >(this)
			, (typename Traits::Path *)0, grammar->path(urule)->asRule());
	Path::ChildState res = Path::eReady;
	//while(res != PathT::eComplete) {
	for(int n = 0; (res != Path::eMatched) && (n < 3); n++) {
		SemanticValue v = scanner.read();
		tokens.push_back(v);
		// Need to uncover "process()" in PathT<Traits> as it seems to be hidden!
		//res = root.PathT<Traits>::process();
		res = root.process();
		switch(res) {
		case Path::eReady:
		case Path::eMatched:
			break;
		case Path::eFailed:
			throw runtime_error("Parse failed.");
		}
		dump();
	}
	return 0;
}
template< typename Traits >
inline bool ParserT< Traits >::isToken(unsigned int symbol) const
{
	return symbol < grammar->first_rule_index;
}
template< typename Traits >
inline void ParserT< Traits >::stream(ostream & s_) const
{
	TreeStreambuf sb(s_);
	stream(sb);
}
template< typename Traits >
inline void ParserT< Traits >::stream(TreeStreambuf & sb) const
{
	sb.add("tokens", true);
	sb.add("======", true);
	sb.push();
	typename vector < SemanticValue >::const_iterator i = tokens.begin();
	typename vector < SemanticValue >::const_iterator j = tokens.end();
	unsigned int last = tokens.size() - 1;
	for(unsigned int n = 0; i != j; ++i, ++n)
		static_cast< const typename Traits::Parser * >(this)->stream(sb, *i, 0, n, n == last);
	sb.pop();
	sb.add("paths", true);
	sb.add("=====", true);
	root.stream(sb);
}
template< typename Traits >
inline void ParserT< Traits >::dump() const
{
#if defined(V3C_TRACE)
	//stringstream ss;
	//stream(ss);
	//trace_raw << ss.str() << endl;
	cout << '\n';
	{
		PrefixStreambuf c(cout);
		stream(cout);
	}
	cout << '\n';
#endif
}
template< typename Traits >
inline unsigned int ParserT< Traits >::next_path_id()
{
	return m_next_path_id++;
}
} // namespace runtime

inline const runtime::Node * Grammar::path(unsigned int num) const
{
	if(num < first_rule_index)
		throw runtime_error(string_inline_printf("Grammar::rule : %u < %u"
			, num, first_rule_index));
	const runtime::Node * m_pNode = paths[num - first_rule_index];
	return m_pNode;
}
inline const char * Grammar::name(unsigned int symbol) const
{
	return symbol_names[symbol - runtime::Path::eFirstTokenIndex];
}

} // namespace bisonx
} // namespace v3c
