// Copyright (C) 2009  Philip Ashmore (contact@philipashmore.com)
// License: LPGLv3.  See LICENSE.txt for the full license.

/*
The purpose of this test is to test the bisonx runtime implementation with the
calc-x calculator scanner/grammar.

Of course bisonx isn't complete yet so we kick things off with a no-op grammar.

To do this it needs to access several namespaces
 -# bisonx::runtime
	the classes that implement the parser
 -# bisonx::calc
	the semantic value type passed from the scanner to the parser, as well as
	the scanner itself

This simple example serves as the starting point for the bisonx-generated code
that implements the parser.
Alternatively bisonx could just generate the grammar and the actions and leave
the parser code up to the user - there isn't that much of it.
*/

enum
{ rUnit        = 303
, rUnit0       = 304
, rExp         = 305
, rAssignments = 306
, rAssignment  = 307
, rAssignment0 = 308
, rExp0        = 309
, rExp1        = 310
, rExp2        = 311
, rExp3        = 312
, rExp4        = 313
, rExp5        = 314
};

#include <v3c/trace_scope.hpp>
#include <map>
#include "parser.hpp"
#include "calc_scanner.hpp"
#include <v3c/i_chunk_stream.hpp>

using namespace std;
using namespace v3c;
using namespace bisonx;

/*!
Here's a cut-down version of the calculator grammar.

%token        END          0 "end of file"
%token        IDENTIFIER 300 "identifier"
%token        ASSIGN     301 ":="
%token        NUMBER     302 "number"

%left '+' '-';
%left '*' '/'; // Precedence greater than +,-.
// (0)
unit: assignments exp {A} // The entry point.
;
// I'm breaking the notation here - B sets up the accumulator for the result.
assignments: assignments<B> assignment {C}
           | / * Nothing.  * /         {D}
;
assignment: "identifier" ":=" exp {E}
; // MergeFunc is F
exp: exp '+' exp  {G}
   | exp '-' exp  {H}
   | exp '*' exp  {I}
   | exp '/' exp  {J}
   | "identifier" {K}
   | "number"     {L}
;
*/
const char * action_name[] =
{ "unit: assignments exp (A)"
, "assignments: assignments{B} assignment"
, "assignments: assignments assignment {C}"
, "assignments: assignments assignment | /* Nothing. */ {D}"
, "assignment: \"identifier\" \":=\" exp {E}"
, "exp <merge> {F}"
, "exp '+' exp {G}"
, "exp '-' exp {H}"
, "exp '*' exp {I}"
, "exp '/' exp {J}"
, "\"identifier\" {K}"
, "\"number\" {L}"
};
////////////////////////////////////////////////////////////////////////////////
// bisonx-generated code begins
enum Rules { unit=303, assignments, assignment, exp };

extern "C" {
	extern const Grammar bisonx_calc_grammar;
}

// bisonx-generated code ends
////////////////////////////////////////////////////////////////////////////////

class RuleToken : public calc::Token
{
public:
	enum { eRule = calc::Token::eNext, eNext };
	/// Called directly
	inline RuleToken(int id_)
	: calc::Token(eRule, id_, v3c::Span())
	, items()
	{
	}
	virtual inline const char * name() const
	{
		if(type() != eRule)
			return Token::name();
		return "rule";
	}
	vector < calc::TokenPtr > items;
private:
	RuleToken(const RuleToken &);
	RuleToken & operator =(const RuleToken &);
};

struct Traits;
class Parser;

struct Traits
{
	typedef class Parser Parser;
	typedef calc::TokenPtr SemanticValue;
	typedef calc::Scanner Scanner;
	typedef runtime::PathT< Traits > Path;
	typedef smart_ptr< Path, SingleThreadModelTraits > PathPtr;
	typedef vector< PathPtr > SubPaths;
};
class Parser : public runtime::ParserT< Traits >
{
public:
	typedef runtime::ParserT< Traits > Base;
	enum {A,B,C,D,E,F,G,H,I,J,K,L};

	inline Parser(const Grammar & grammar_)
	: Base(grammar_)
	{
	}
	runtime::Path::ChildState onAction(Traits::Path & path, int16_t action)
	{
		const char * name = "<default>";
		if(action != -1)
			name = action_name[action];
		TraceWrap("### Parser::onAction(" << name << ") ###" );
		switch(action) {
		case A:
			break;
		case B:
			path.asRepeatRule().value() = new RuleToken(rAssignments);
			break;
			//path.value_stack.push_back(new RuleToken(rExp));
		case C: {
			Traits::Path::RepeatRulePath & p = path.asRepeatRule();
			calc::TokenPtr & v = p.value();
			if(v->type() != RuleToken::eRule)
				throw runtime_error("Wrong type for assignments token!");
			RuleToken * r = static_cast< RuleToken * >(v.pointer());
			r->items.push_back(p.m_pRule->value());
			break; }
		case D:
			break;
		case E: {
			RuleToken * r = new RuleToken(rAssignment);
			Traits::Path::ProductionPath & p = path.asProduction();
			r->items = p.m_value_stack;
			p.m_value_stack.clear();
			p.m_value_stack.push_back(r);
			break; }
		case G:
		case H:
		case I:
		case J:
		case K:
			break;
		case L: {
			Traits::Path::ProductionPath & p = path.asProduction();
			trace << "Got " << str(p.value()) << '.' << endl;
			RuleToken * r = new RuleToken(rExp);
			r->items.push_back(p.value());
			path.asProduction().value() = r;
			break; }
		default:
			// $$ = $1
			break;
		}
		return runtime::Path::eReady;
	}
	runtime::Path::ChildState onMerge(Traits::Path &path, int16_t func)
	{
		const char * name = "<default>";
		if(func != -1)
			name = action_name[func];
		TraceWrap("### Parser::onMerge(" << name << ") ###" );
		switch(func) {
		case F:
		case -1: {
			// $$ = $1
			Traits::Path::RulePath & rule = path.asRule();
			rule.value() = rule.want_merge[0]->value();
			break; }
		default:
			break;
		}
		return runtime::Path::eMatched;
	}
	/*!
	exp
	|- exp
	|   |- exp
	|   |  \- a
	|   |- *
	|   \- exp
	|      \- b
	|- +
	\- exp
		\- c
	*/
	void stream(TreeStreambuf & sb, const calc::TokenPtr & p
	, const RuleToken * /*parent*/, unsigned int n, bool last) const
	{
		stringstream ss;
		ss << '[' << setw(2) << n << "] ";
		if(p) switch(p->type()) {
		case RuleToken::eRule: {
			const RuleToken * r = static_cast< const RuleToken * >(&*p);
			ss << grammar->name(r->id());
			sb.add(ss.str(), last);
			sb.push();
			vector < calc::TokenPtr >::const_iterator i = r->items.begin();
			vector < calc::TokenPtr >::const_iterator j = r->items.end();
			unsigned int ulast = r->items.size() - 1;
			for(unsigned int m = 0; i != j; ++i, ++m)
				stream(sb, *i, r, m, m == ulast);
			sb.pop();
			return; }
		default:
			ss << str(p);
		}else{
			ss << "<null>";
		}
		sb.add(ss.str(), last);
	}
	string str(const calc::TokenPtr & p) const
	{
		stringstream ss;
		if(p) switch(p->type()) {
		case RuleToken::eRule:
			ss << "Rule " << grammar->name(p->id());
			break;
		case calc::Token::eLex:
			if(!p->id())
				ss << "<EOF>";
			if(p->id() < runtime::Path::eFirstTokenIndex)
				ss << '\'' << (char)p->id() << '\'';
			else
				ss << grammar->name(p->id());
			break;
		case calc::Token::eInt:
			ss << p->int_value();
			break;
		case calc::Token::eString:
			ss << p->string_value();
			break;
		default:
			ss << "?";
		}else{
			ss << "<null>";
		}
		return ss.str();
	}
};

int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	CoutTraceWriter ctw("");

	TraceWrap(__PRETTY_FUNCTION__);

	Parser p(bisonx_calc_grammar);
	const char * str = "a := 1 b := 2 c := 3 d := a + b * c d";
	i_chunk_stream is(str, str + strlen(str));
	p.scanner.set_debug(true);
	p.scanner.switch_streams(&is, (ostream *)0);
	int res = p.parse();

	trace_cout << (res ? "Failure" : "Success!") << endl;
	return res;
}
