// Copyright (C) 2009  Philip Ashmore (contact@philipashmore.com)
// License: LGPLv3.  See LICENSE.txt for the full license.

/** \file
 * \brief Header for v3c::Offset, v3c::Position and v3c::Span
 */
#if !defined(v3c_span_hpp_included)
  #define v3c_span_hpp_included

  #include <iostream>
  #include <sstream>

namespace v3c
{

using namespace std;

struct Offset
{
	inline Offset(int lines_ = 0, int cols_ = 0)
	: lines(lines_)
	, cols(cols_)
	{
	}
	inline Offset(const char * text, unsigned int length)
	: lines()
	, cols()
	{
		init(text, length);
	}
	inline Offset(const string & text)
	: lines()
	, cols()
	{
		unsigned int length = text.length();
		init(text.c_str(), length);
	}
	inline void init(const char * text, unsigned int length)
	{
		lines = cols = 0;
		for(unsigned int n = 0; n < length; n++)
			switch(text[n]) {
			case '\n':
				lines++;
				cols = 0;
				break;
			case '\t':
				cols += 4; // TODO tls for tab-width?
				break;
			case '\r':
				break;
			default:
				cols++;
				break;
			}
	}
	int lines, cols;
};

struct Position
{
	Position
	( unsigned int line_ = 0
	, unsigned int column_ = 0
	, const string & filename_ = "")
	: line(line_)
	, column(column_)
	, filename(filename_)
	{
	}
	unsigned int line, column; // 1-based, 0 == not set
	string filename;
	void newline() { line++; column = 1; }
	void add_columns(unsigned int n) { column += n; }
	void update_position(const char * text, unsigned int length)
	{
		Offset offs(text, length);
		*this += offs;
	}
	void new_file(const string & filename_)
	{
		filename = filename_;
		line = column = 1;
	}

	Position operator +(const Offset & o) const
	{
		if(o.lines)
			return Position(line + o.lines, o.cols + 1, filename);
		return Position(line, column + o.cols, filename);
	}
	Position operator -(const Offset & o) const
	{
		return Position(line - o.lines, column - o.cols, filename);
	}
	Position & operator +=(const Offset & o)
	{
		if(o.lines)
			line += o.lines, column = o.cols + 1;
		else
			column += o.cols;
		return *this;
	}
	Position & operator -=(const Offset & o)
	{
		line -= o.lines, column -= o.cols;
		return *this;
	}

	string opt_filename() const
	{
		if(filename.length())
			return filename + ", ";
		return filename;
	}
	ostream & stream_position(ostream & s) const
	{
		s << opt_filename()
			<< "line " << line << ", col " << column;
		return s;
	}
};
inline
ostream & operator <<(ostream & s, const Position & l)
{
	return l.stream_position(s);
}

struct Span
{
	Span
	( int begin_line_
	, int begin_column_
	, const string & begin_filename_
	, int end_line_
	, int end_column_
	, const string & end_filename_)
	: begin(begin_line_, begin_column_, begin_filename_)
	, end(end_line_, end_column_, end_filename_)
	{
	}
	Span()
	: begin()
	, end()
	{
	}
	Span(const Position & l, unsigned int cols, unsigned int lines = 0)
	: begin(l)
	, end(l + Offset(lines, cols))
	{
	}
	Span(const Position & l, const string & s)
	: begin(l)
	, end(l + Offset(int(0), (int)s.length()))
	{
	}
	Span(const Position & b, const Position & e)
	: begin(b)
	, end(e)
	{
	}
	Position begin, end; // Like iterators - if begin == end the span is empty.

	void newline() { end.line++; end.column = 1; }
	void add_columns(unsigned int n) { end.column += n; }
	void zero() { begin.line = end.line, begin.column = end.column; }

	void reset(const Position & l, unsigned int length)
	{
		begin.line = end.line = l.line, begin.column = l.column;
		end.column = l.column + length;
		end.filename = l.filename;
	}
	void update(const Span & l)
	{
		if(begin.line == 0) {
			begin.line = l.begin.line;
			begin.column = l.begin.column;
			begin.filename = l.begin.filename;
		}
		if(end.line < l.end.line) {
			end.line = l.end.line;
			end.column = l.end.column;
		}else if(end.column < l.end.column) {
			end.column = l.end.column;
		}
		end.filename = l.end.filename;
	}
	ostream & stream_span(ostream & s) const
	{
		if(begin.filename == end.filename) {
			if(begin.line == end.line)
				s << begin.opt_filename()
					<< "line " << begin.line << ", col " << begin.column
					<< " : " << end.column - begin.column;
			else
				s << begin << " - line " << end.line << ", col " << end.column;
		}else{
			s << begin << " -> " << end;
		}
		return s;
	}
	string string_span() const
	{
		stringstream ss;
		stream_span(ss);
		return ss.str();
	}
	unsigned int column_width() const
	{
		return end.column - begin.column;
	}
};
inline
ostream & operator <<(ostream & s, const Span & l)
{
	return l.stream_span(s);
}

} // namespace v3c

#endif // #if !defined(v3c_span_hpp_included)
